/*
 * wmdimp
 * Copyright (c) 2004 Daniele Tamino <aslo@wmdimp.cjb.net>.
 * This software covered by the GPL.  See COPYING file for details.
 *
 * debug.c
 *
 * This file contains utilities useful for debugging
 *
 */
#ifndef NDEBUG

#define _GNU_SOURCE


#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include "debug.h"

#if DBG_MESSAGES

#include <stdarg.h>

static char* progname; /* = argv[0] */
enum { WAIT_SIGNAL = SIGUSR2 };

void do_wait(int n)
{
    pid_t me = getpid();
    fprintf(stderr, "*** %d: waiting %d seconds ***\n", me, n);
    while((n = sleep(n)) > 0);
    fprintf(stderr, "*** %d: done waiting ***\n", me);
}

static pid_t the_one = 0;
void wait_handler(int sig)
{
    (void)sig;
    if(getpid() == the_one)
    do_wait(5); /* the crashing thread will get up first */
    else
    do_wait(10);
}

void __dbg_start_debugger(void)
{
    pid_t myself = getpid();
    pid_t child;
    if((child=fork()) == 0) {
    char myself_str[10];
    setpgid(0, getpid());
    fprintf(stderr, "*** running debugger ***\n");
    snprintf(myself_str, sizeof(myself_str), "%d", (int)myself);
    if(getenv("DISPLAY"))
        execlp("xterm", "xterm", "-e", "/usr/bin/gdb", progname,
           myself_str, NULL);
    else
        /* this doesn't still work reliably :( */
        execlp("/usr/bin/gdb", "gdb", progname, myself_str, NULL);
    exit(-1);
    }
    setpgid(child, child);
}

void errsig_handler(int sig)
{
    /* io speriamo che me la cavi */
    int mygrp = getpgrp();
    the_one = getpid();
    fprintf(stderr, "wmdimp(%d): %s\nRunning debugger...\n",
        getpid(), strsignal(sig));
    __dbg_start_debugger();
    kill(-mygrp, WAIT_SIGNAL);
    signal(sig, SIG_DFL); /* we started the debugger, it will take
                 care of this */
}

void dbg_start_debugger(void)
{
    int mygrp = getpgrp();
    the_one = getpid();
    __dbg_start_debugger();
    kill(-mygrp, WAIT_SIGNAL);
}

void dbg_init(int argc, char** argv)
{
    (void) argc;
    progname = strdup(argv[0]);
    /* mo' ce proviamo... */
    signal(SIGSEGV, errsig_handler);
    signal(SIGILL, errsig_handler);
    signal(SIGFPE, errsig_handler);
    signal(SIGUSR2, wait_handler);
#if DBG_MEMORY
    mtrace();
    atexit(muntrace);
#endif
}

void print_time(FILE* out, struct timeval* t)
{
    double sec = (t->tv_sec % 60) + (double)(t->tv_usec)/1e6;
    int min  = t->tv_sec / 60;
    int hour = min / 60;
    min %= 60;
    hour %= 24; /* I don't mind exact time, just intervals */
    if(hour > 0) fprintf(out, "%dh", hour);
    if(min > 0)  fprintf(out, "%dm", min);
    fprintf(out, "%.3fs", sec);
}

int dbg_printf(const char* file, int line, const char* fmt, ...)
{
    struct timeval tv;
    int ret;
    va_list va;
    va_start(va, fmt);
    gettimeofday(&tv, NULL);
    fprintf(stderr, "DBG(%d,", getpid());
    print_time(stderr, &tv);
    fprintf(stderr, ") %s:%d: ", file, line);
    ret = vfprintf(stderr, fmt, va);
    va_end(va);
    return ret;
}

void dbg_dump_buffer(const unsigned char* buf, int size)
{
    int i;
    const unsigned char *p;
    for(p = buf; size >= 16; size -= 16, p +=16) {
    fprintf(stderr, "%08X ", p-buf);
    for(i = 0; i < 8; i++)
        fprintf(stderr, "%02X ", (int)p[i]);
    fprintf(stderr, " ");
    for(; i < 16; i++)
        fprintf(stderr, "%02X ", (int)p[i]);
    fprintf(stderr, "|");
    for(i = 0; i < 16; i++)    
        fprintf(stderr, "%c", isprint(p[i]) ? p[i] : '.');
    fprintf(stderr, "|\n");
    }
    if(size > 0) {
    fprintf(stderr, "%08X ", p-buf);
    for(i = 0; i < size; i++)
        if(i < size) fprintf(stderr, "%02X ", (int)p[i]);
        else     fprintf(stderr, "   ");
    fprintf(stderr, " ");
    for(; i < 16; i++)  
        if(i < size) fprintf(stderr, "%02X ", (int)p[i]);
        else     fprintf(stderr, "   ");
    fprintf(stderr, "|");
    for(i = 0; i < 16; i++)     
        if(i < size) fprintf(stderr, "%c", isprint(p[i]) ? p[i] : '.');
        else     fprintf(stderr, " ");
    
    fprintf(stderr, "|\n");
    }
}

#endif /* DBG_MESSAGES */

#if DBG_METER

/* debug meter */
#include <X11/Xlib.h>
#include "wmgeneral.h"

typedef struct dbg_meter {
    Window win;
    GC gc;
    Pixel fg, bg;
} dbg_meter;

void* dbg_meter_new(const char* name)
{    
    int screen;
    XFontStruct *fnt;
    XGCValues values;
    dbg_meter* m = (dbg_meter*)malloc(sizeof(dbg_meter)); 
    screen = DefaultScreen(display);  
    m->bg = WhitePixel(display, screen);
    m->fg = BlackPixel(display, screen);

    if(!m) { DBGMSG("dbg_meter_new: OUT OF MEMORY!\n"); return NULL; }
    XLockDisplay(display);
    m->win = XCreateSimpleWindow(display, Root, 0, 0, 100, 20, 2, m->bg, m->bg);
    m->gc = XCreateGC(display, win, 0, &values);
    XSetForeground(display, m->gc, m->fg);
    XSetBackground(display, m->gc, m->bg);
    fnt = XLoadQueryFont(display, "9x15");
    XSetFont(display, m->gc, fnt->fid);
    XMapWindow(display, m->win);
    XStoreName(display, m->win, name);
    XUnlockDisplay(display);
    return m;
}

void dbg_meter_show(void* id, int percent)
{
    char str[20];
    int len;
    dbg_meter* m = (dbg_meter*) id;
    if(!m) return;
    /*len = sprintf(str, "%d%%", percent);*/
    XLockDisplay(display);
    XSetForeground(display, m->gc, m->bg);
    XFillRectangle(display, m->win, m->gc, percent, 0, 100-percent, 20);
    XSetForeground(display, m->gc, m->fg);
    XFillRectangle(display, m->win, m->gc, 0, 0, percent, 20);
    /*XDrawString(display, m->win, m->gc, 20, 20, str, len);*/
    XFlush(display);
    XUnlockDisplay(display);
}

void dbg_meter_delete(void* id)
{
    dbg_meter* m = (dbg_meter*) id;
    if(!m) return;
    XLockDisplay(display);
    XFreeGC(display, m->gc);
    XDestroyWindow(display, m->win);
    XUnlockDisplay(display);
    free(m);
}

#endif /* DBG_METER */

#endif /* NDEBUG */
