/*
 * wmdimp
 * Copyright (c) 2004 Daniele Tamino <aslo@wmdimp.cjb.net>
 *
 * Based on wmmp3: Copyright (c)1999 Patrick Crosby <xb@dotfiles.com>.
 *
 * This software covered by the GPL.  See COPYING file for details.
 *
 * mpg123_player.c
 *
 * player that uses "mpg123 -R dummy" to play mp3's
 *
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include <errno.h>
#include <stdarg.h>
#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "mpg123_player.h"
#include "utils.h"
#include "timer.h"
#include "debug.h"

int mpg123_static_init(opt_hash* ht);
void mpg123_static_finish(void);
int mpg123_init(mpg123_player*);
player* mpg123_create(void);

void mpg123_finish(struct player*);
int mpg123_test(struct player*, const char* fn);
int mpg123_setup(struct player*, const char* fn, char ** title, int play);
int mpg123_play(struct player*);
int mpg123_play_sync(struct player*);
int mpg123_pause(struct player*);
int mpg123_stop(struct player*);
int mpg123_jump_frames(struct player*, long f);
int mpg123_jump_time(struct player*, long sec);
int mpg123_jump(struct player*, long position, long total);
int mpg123_get_length(struct player*, song_pos* len);
int mpg123_get_pos(struct player*, song_pos* pos);
void mpg123_set_callback(struct player*, player_cb_func cb, void* param);

/* template to initialize a new struct with */
player mpg123_player_tmpl = {
    "mp3 player (using mpg123)",
    st_stop,
    /* methods */
    mpg123_finish,
    mpg123_test,
    mpg123_setup,
    mpg123_play,
    mpg123_play_sync,
    mpg123_pause,
    mpg123_stop,
    mpg123_jump_frames,
    mpg123_jump_time,
    mpg123_jump,
    mpg123_get_length,
    mpg123_get_pos,
    mpg123_set_callback
};

player* mpg123_create(void) {
    mpg123_player* pl = (mpg123_player*) malloc(sizeof(mpg123_player));
    if(!pl) return NULL;
    mpg123_init(pl);
    return &pl->pl;
}

player_factory mpg123_player_factory = {
    mpg123_static_init,
    mpg123_static_finish,
    mpg123_create,
    {
    "mpg123 player",
    /* path to mpg123 executable (searched in $PATH)*/
    "mpg123 = string:mpg123;\n"
    }
};

static char** mpg123_argv = NULL;

/*
 * functions that actually control mpg123
 */

/* send a command to mpg123 */
int cmd_printf(mpg123_player* me, char* fmt, ...)
{
    va_list vl;
    int ret = 0;
    if(me->play_pid > 0) {
        va_start(vl, fmt);
        ret = vfprintf(me->play_wr, fmt, vl);   
        va_end(fmt);
        fflush(me->play_wr);
    }
    return ret;
}

mpg123_player* sighnd_me = NULL; /* used by signal handler */

void close_double_pipe(mpg123_player* me);

void shutdown_mpg123(mpg123_player* me, int error)
{
    int status;
    sighnd_me = NULL;
    signal(SIGCHLD, SIG_IGN);
    if(!error) {
        DBGMSG("sending QUIT...\n");
        cmd_printf(me, "QUIT\n");
        usleep(1000); /* give it a little time to shut down ... */
    }
    if(me->play_pid && waitpid(me->play_pid, &status, WNOHANG) == 0) {
        /* this way it should terminate immediately */
        DBGMSG("killing mpg123...\n");
        kill(me->play_pid, SIGINT);
        kill(me->play_pid, SIGTERM);
        kill(me->play_pid, SIGKILL);
        waitpid(me->play_pid, &status, 0);
    }
    DBGMSG("mpg123 closed (%d), closing down streams...\n", status);
    /* close all pipe streams */
    me->play_pid = 0;
    close_double_pipe(me);
    /* update play status */
    me->curr_status = me->pl.status = st_stop;
    me->curr_pos.frames = 0;
    me->curr_pos.seconds = 0.0;
    pthread_cond_broadcast(&me->play_cond);
    if(me->callback)
        me->callback(&me->pl, CB_STATUS, me->cb_param, NULL);
}

/*
 * Main play thread (receives messages from mpg123)
 */

void play_thread_cancel(void* param)
{
    mpg123_player* me = (mpg123_player*) param;
    shutdown_mpg123(me, me->play_pid == 0);
}

void* play_thread(void* param)
{
    char buffer[256];
    int done = 0;
    mpg123_player* me = (mpg123_player*) param;

    pthread_cleanup_push(play_thread_cancel, param);
    signal(SIGUSR1, SIG_IGN); /* used to wake up from fgets */

    while(!ferror(me->play_rd) && me->play_pid > 0 && !done) {
        char* ret;
        errno = 0;
        /* use PTHREAD_CANCEL_ASYNCHRONOUS to avoid blocking if cancelled */
        pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
        ret = fgets(buffer, sizeof(buffer), me->play_rd);
        /* ok, for the rest, deferred mode is better */
        pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

        if(ret == NULL) {
            /* especially, we need to catch EPIPE */
            if(errno != 0 && errno!= EINTR) {
                me->error = 1;
                DBGMSG("play_thread: fgets returned errno=%d (%s)\n",
                       errno, strerror(errno));
                if(me->callback)
                    me->callback(&me->pl, CB_ERROR, me->cb_param, strerror(errno));
                break;
            } else
                continue;
        }

        DBG(if(buffer[1] != 'F')
            DBGMSG(buffer);
        )

        if(buffer[0] == '@') {
            switch(buffer[1]) {
            case 'R': /* tag 'MPG123' */
                if(strncmp(buffer+3, "MPG123", 6) != 0)
                    me->error = 1;
                else {
                    me->error = 0;
                    DBGMSG("signalling main thread...\n");
                    pthread_cond_broadcast(&me->play_cond);
                }
                break;
            case 'I': /* nome file / ID3 */
                if(strncmp(buffer+3, "ID3:", 4) == 0) {
                    int len;
                    pthread_cleanup_push((cleanup_f)pthread_mutex_unlock,
                                         (void *) &me->title_mtx);
                    pthread_mutex_lock(&me->title_mtx);

                    /* alloc always the maximum size, and reuse old memory */
                    if(!me->song_title)
                        me->song_title = malloc(64);
                    if(me->song_title) {
                        char* title = me->song_title;
                        /* use a simple 'Title - Artist' display */
                        memcpy(title, buffer + 7, 30); /* get title */
                        title[30]='\0';
                        stripspace(title);
                        strcat(title, " - ");
                        len = strlen(title);
                        memcpy(title + len, buffer + 37, 30); /* get artist */
                        title[len + 30] = '\0';
                        stripspace(title);
                    }

                    pthread_cleanup_pop(1); /* unlock the mutex */
                } else {
                    int len = strlen(buffer);
                    buffer[len - 1] = '\0'; /* remove '\n' */
                    me->song_title = malloc(len - 3);
                    if(!me->song_title) {
                        fprintf(stderr, "VERY LOW MEMORY!!!\n");
                        me->song_title = me->filename;
                    } else {
                        memcpy(me->song_title, buffer + 3, len - 3);
                    }
                }
                DBGMSG("play_thread: song title set to '%s'\n", me->song_title);
                /* use CB_TITCHANGE to show the new title */
                if(me->callback)
                    me->callback(&me->pl, CB_TITCHANGE, me->cb_param,
                                 me->song_title);
                me->song_len.frames = -1;
                me->curr_status = st_play;
                pthread_cond_broadcast(&me->play_cond);
                break;
            case 'S': /* file info */
                /* ignored */
                break;
            case 'F': { /* status updates - very frequent */
                /* Note: mpg123 (but not mpg321), loading a new song without
                   stopping the previous, sends a few @F lines before changing.
                   This is a problem since we change me->curr_pos and last_pos in
                   play() to zero, but this fakes up everything.
                   To avoid the problem, we check if last_pos > me->curr_pos
                   (see player.c)
                */
                int frames, fr_remain;
                float secs, sec_remain;
                char* s;
                pthread_cleanup_push((cleanup_f)pthread_mutex_unlock,
                                     (void *) &me->curr_pos_mtx);
                pthread_mutex_lock(&me->curr_pos_mtx);

                s = buffer + 3;
                frames = strtol(s, &s, 0);
                fr_remain = strtol(s, &s, 0);
                secs = strtod(s, &s);
                sec_remain = strtod(s, &s);
                
                if(me->song_len.frames < 0) {
                    me->song_len.frames = frames + fr_remain;
                    me->song_len.seconds = secs + sec_remain;
                    pthread_cond_broadcast(&me->play_cond);
                }
                me->curr_pos.frames = frames;
                me->curr_pos.seconds = secs;
                pthread_cleanup_pop(1);

                if(me->callback)
                    me->callback(&me->pl, CB_UPDATE, me->cb_param, NULL);
                break;
            }
            case 'P': /* stop/pause status */
                sscanf(buffer+3, "%d", &me->curr_status);
                DBGMSG("new status = %d, wanted status = %d\n",
                       me->curr_status, me->wanted_status);
                if(me->curr_status != me->wanted_status) {
                    /* status changed without our will */
                    if(me->curr_status==st_stop && me->wanted_status==st_play) {
                        if(me->sync) done = 1;
                        else
                        if(me->callback)
                            me->callback(&me->pl, CB_SONGCHANGE, me->cb_param, NULL);
                    } else {
                        me->pl.status = me->curr_status;
                        if(me->callback)
                            me->callback(&me->pl, CB_UPDATE, me->cb_param, NULL);
                    }
                    /* I don't think this will happen in other cases */
                }
                if(me->curr_status == st_stop) {
                    pthread_cleanup_push((cleanup_f)pthread_mutex_unlock,
                                         (void *) &me->curr_pos_mtx);
                    pthread_mutex_lock(&me->curr_pos_mtx);
                    me->curr_pos.frames = 0;
                    me->curr_pos.seconds = 0.0;
                    pthread_cleanup_pop(1);
                }
                pthread_cond_broadcast(&me->play_cond);
                break;
            case 'E': /* error message */
                if(me->callback)
                    me->callback(&me->pl, CB_ERROR, me->cb_param, buffer + 3);
                me->error = 1;
                break;
            } /* end switch(output type) */
        } else {
            if(me->callback)
                me->callback(&me->pl, CB_ERROR, me->cb_param, buffer);
        }
    } /* end main thread loop */
    
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
    DBGMSG("play_thread: cleaning up...\n");
    pthread_cleanup_pop(1); /* execute cleanup */
    DBGMSG("play_thread: bye!\n");
    return 0; 
}

void terminate_play_thread(mpg123_player* me)
{
    if(me->play_thr != (pthread_t)0) {
        DBGMSG("terminating play thread...\n");
        pthread_cancel(me->play_thr);
        pthread_join(me->play_thr, NULL);
        me->play_thr = (pthread_t)0;
        DBGMSG("play thread terminated!\n");
    }
}

/* Forks a new mpg123 process redirecting its stdin and stdout
 * to me->play_wr and me->play_rd, respectively.
 * Returns 1 (true) if success, 0 (false) otherwise.
 */
int open_double_pipe(mpg123_player* me)
{
    int mypiperd[2];
    int mypipewr[2];
    /* open pipes */
    if(pipe(mypiperd)<0 || pipe(mypipewr)<0) {
        perror("pipe");
        return 0;
    }

    /* fork the new mpg123 process */
    DBGMSG("forking mpg123...\n");
    me->play_pid = fork();
    if(me->play_pid == 0) { /* child process? */
        /* set standard input, output and error to our pipes */
        close(mypipewr[1]);
        close(mypiperd[0]);

        dup2(mypipewr[0], STDIN_FILENO);
        dup2(mypiperd[1], STDOUT_FILENO);
        dup2(mypiperd[1], STDERR_FILENO); /* stderr --> stdout (sh: 2>&1) */
        close(mypipewr[0]);
        close(mypiperd[1]);

        /* exec mpg123 with -R option (remote control) */
        execvp(mpg123_argv[0], mpg123_argv);
        fprintf(stderr, "cannot execlp %s: %s\n",
                mpg123_argv[0], strerror(errno));
        exit(-1);
    }
    if(me->play_pid < 0) {
        perror("cannot fork");
        return 0;
    }
    /* setup pipes */
    close(mypiperd[1]);
    close(mypipewr[0]);
    me->play_rd = fdopen(mypiperd[0], "r");
    if(!me->play_rd) { perror("fdopen for reading"); return 0; }
    me->play_wr = fdopen(mypipewr[1], "w");
    if(!me->play_wr) { perror("fdopen for writing"); return 0; }

    setvbuf(me->play_rd, NULL, _IOLBF, BUFSIZ); /* set line buffered output */
    setvbuf(me->play_wr, NULL, _IOLBF, BUFSIZ);

    return 1; /* tutto ok! */
}

void close_double_pipe(mpg123_player* me)
{
    if(me->play_rd != NULL && me->play_wr != NULL) {
        int rdfd = fileno(me->play_rd);
        int wrfd = fileno(me->play_wr);
        fclose(me->play_rd);
        fclose(me->play_wr);
        me->play_rd = me->play_wr = NULL;
        close(rdfd);
        close(wrfd);
    }
}

void signal_child() 
{
    int status;
    DBGMSG("received SIGCHLD... \n");
    if(sighnd_me->play_pid == 0 ||
       waitpid(sighnd_me->play_pid, &status, WNOHANG) != 0) {
        DBGMSG("  (mpg123 status=%d), stopping play_thread...\n", status);
        sighnd_me->error = 1;
        /* needed before terminate_play_thread to avoid SIGPIPE */
        sighnd_me->play_pid = 0;
        sighnd_me->curr_status = st_stop;
        if(sighnd_me->pl.status != st_stop) {
            sighnd_me->pl.status = st_stop;
            if(sighnd_me->callback)
                sighnd_me->callback(&sighnd_me->pl, CB_STATUS,
                                    sighnd_me->cb_param, NULL);
        }
        /*terminate_play_thread(sighnd_me);*/
        /* it will take care by itself to terminate */
        pthread_kill(sighnd_me->play_thr, SIGUSR1); /* wake up from fgets */
    } else {
        DBGMSG("  but mpg123 is not dead!!! (ignored)\n");
    }
}

#define ASSERT_MALLOC(x) if(!(x)) { fprintf(stderr, "BUY MORE RAMM"); return; }

void free_mpg123_argv(void)
{
    if(mpg123_argv) {
        char** s;
        for(s = mpg123_argv; *s != NULL; s++)
            free(*s);
        free(mpg123_argv);
        mpg123_argv = NULL;
    }
}

void set_mpg123(char *cmd)
{
    struct vect argv;
    char* s;
    char** ps;
    free_mpg123_argv();
    ASSERT_MALLOC(vect_init(&argv, sizeof(char*)));
    cmd = skipspaces(cmd);
    while(*cmd != '\0') {
        s = read_string(cmd, &cmd, NULL);
        cmd = skipspaces(cmd);
        ASSERT_MALLOC(s);
        ASSERT_MALLOC(vect_incr(&argv, 1));
        *_vect_get_last(&argv, char*) = s;
    }
    ASSERT_MALLOC(vect_incr(&argv, 3));
    ps = _vect_get_last(&argv, char*);
    *(ps - 2) = strdup("-R");
    *(ps - 1) = strdup("dummy"); /* required */
    *ps = NULL; /* terminator */
    mpg123_argv = (char**) argv.v;
}

/* initialize static structures (loads vaules from config) */
int mpg123_static_init(opt_hash* ht)
{
    opt_tree* t = opt_hash_find(ht, "mpg123");
    assert(t && t->root && t->root->next);
    set_mpg123(t->root->next->val.s);
    return 1;
}

/* releases static structures */
void mpg123_static_finish(void)
{   
    free_mpg123_argv();
}

int mpg123_init(mpg123_player *me) /* constructor */
{
    DBGMSG("mpg123_init(%p) called from thr %ld.\n",
                me, (long)getpid());
    me->pl = mpg123_player_tmpl; /* copy template with methods */
    me->curr_status = me->wanted_status = st_stop;
    me->error = 0;
    me->play_pid = 0;
    me->play_rd = me->play_wr = NULL;
    me->play_thr = 0;
    me->sync = 0;
    me->filename = NULL;
    me->song_title = NULL;
    me->curr_pos.seconds = 0;
    me->curr_pos.frames = 0;
    me->song_len.seconds = 0;
    me->song_len.frames = -1;
    me->callback = NULL;
    pthread_mutex_init(&me->title_mtx, NULL);
    pthread_mutex_init(&me->curr_pos_mtx, NULL);
    pthread_mutex_init(&me->play_mtx, NULL);
    pthread_cond_init(&me->play_cond, NULL);
    return 1;
}

void mpg123_finish(struct player *pme) /* destructor */
{
    mpg123_player* me = (mpg123_player*) pme;
    DBGMSG("mpg123_finish(%p) called from thr %ld.\n",
                me, (long)getpid());

    terminate_play_thread(me);

    if(me->song_title) {
        if(me->song_title != me->filename)
            free(me->song_title);
        me->song_title = NULL;
    }
    if(me->filename) {
        free(me->filename);
        me->filename = NULL;
    }

    pthread_cond_destroy(&me->play_cond);
    pthread_mutex_destroy(&me->play_mtx);
    pthread_mutex_destroy(&me->curr_pos_mtx);
    pthread_mutex_destroy(&me->title_mtx);
}

enum { WAIT_FOR_MPG123_USEC = 1000000 /* 1 sec of max waiting */ };

int start_play_thread(mpg123_player* me)
{
    int ret;
    struct timeval tv;
    struct timespec timeout;

    if(me->play_thr > 0) {
        if(me->play_pid <= 0)
            return 0;
        else
            return 1;
    }
    /* uses mpg123/mpg321 "remote control" interface (-R) */
    if(!open_double_pipe(me))
        return 0;

    /* start play_thread */
    me->error = 1; /* when play_thread receives @R MPG123 it will set it = 0 */
    ret = pthread_create(&me->play_thr, NULL, play_thread, me);
    if(ret != 0) {
        fprintf(stderr, "Unable to create play thread: %s\n",
                strerror(ret));
        shutdown_mpg123(me, 0);
        return 0;
    }
    sighnd_me = me;
    signal(SIGCHLD, signal_child);

    gettimeofday(&tv, NULL);
    /* calculate now + WAIT_FOR_MPG123_USEC */
    tv.tv_usec += WAIT_FOR_MPG123_USEC;
    if(tv.tv_usec > 1000000) {
        tv.tv_sec = tv.tv_sec + tv.tv_usec / 1000000;
        tv.tv_usec %= 1000000;
    }
    timeout.tv_sec = tv.tv_sec;
    timeout.tv_nsec = tv.tv_usec * 1000;

    /* wait for the '@R MPG123' message no more than WAIT_FOR_MPG123_USEC */
    ret = 0;
    pthread_mutex_lock(&me->play_mtx);
    while (me->error == 1 && ret != ETIMEDOUT) {
        ret = pthread_cond_timedwait(&me->play_cond, &me->play_mtx, &timeout);
    } 

    pthread_mutex_unlock(&me->play_mtx);

    if (ret == ETIMEDOUT) {
        fprintf(stderr, "%s is not mpg123 or compatible, "
                "or it does not support -R\n", mpg123_argv[0]);
        shutdown_mpg123(me, 1);
        return 0;
    }

    return 1;
}

/* test if the file is in the correct format for this player.
 * this is just a wrapper around mpg123, how to test if a file is valid?
 * By now, we just load it and stop playing immediately.
 */
int mpg123_test(struct player *pme, const char* fn)
{
    int is_mp3 = 0;
    /* do some checks to see if this is an mp3 file */
    /* this code is taken from xmms-mad (http://xmms-mad.sourceforge.net/)
       (a little bit reformatted and readapted) */
    int32_t check = 0L;
    int fin = open (fn, O_RDONLY);
    (void) pme;
    if (fin >= 0 && read (fin, &check, 4) == 4) {
        /* If first two bytes are a sync header or three bytes are "ID3" */
        if ( ((check & 0xf0ff) == 0xf0ff) || ((check & 0xffffff) == 0x334449)) {
            is_mp3 = 1;
        } else if (check == 0x46464952) { /* RIFF RMP3 stream */    
            lseek (fin, 4, SEEK_CUR);
            if (read (fin, &check, 4) == 4 && check == 0x33504d52)
                is_mp3 = 1;
        }
    }
    /* end xmmd-mad code */
    return is_mp3;    
}

/* setup `filename' for playing and, if current status is st_play,
   start playing it.
   Stores in 'title' (if != NULL) a malloc'ed buffer with the title to show
   (may be something like 'title - album - artist', not just the song title).
   Use me->setup(me, NULL, NULL); to release everything.
*/
int mpg123_setup(struct player *pme, const char* fn, char ** title, int play)
{
    mpg123_player* me = (mpg123_player*) pme;
    int in_play_thread = pthread_equal(pthread_self(), me->play_thr);
    int ret = 0;

    DBGMSG("mpg123_setup(%p, \"%s\", %p) called from thr %ld.\n",
                me, fn, title, (long)getpid());

    if(!fn || !me->filename || strcmp(me->filename, fn) != 0) {
        /* reset song info */
        if(me->song_title) {
            if(me->song_title != me->filename)
                free(me->song_title);
            me->song_title = NULL;
        }
        if(me->filename) {
            free(me->filename);
            me->filename = NULL;
        }
        if(fn)
            me->filename = strdup(fn);
        pthread_mutex_lock(&me->curr_pos_mtx);
        me->song_len.frames = -1;
        me->curr_pos.frames = 0;
        me->curr_pos.seconds = 0;
        pthread_mutex_unlock(&me->curr_pos_mtx);
    }
    if(!fn)
        return 1; /* only a reset was requested */

    if(!play && me->pl.status==st_stop && title==NULL &&
       me->curr_pos.frames >= 0 && me->song_len.frames >= 0) {
        return 1; /* we already have everything */
    }

    /* do not wait for the title, send it later with CB_TITCHANGE */
    me->error = 0;
    me->curr_pos.frames = 0;
    me->curr_pos.seconds = 0.0;
    if(title) *title = NULL;
    if(me->pl.status != st_stop || play) {
        me->pl.status = st_stop; /* so mpg123_play will do its job */
        ret = mpg123_play(pme);
    } else
        ret = 1;

    DBGMSG("mpg123_setup: done!\n");
    return ret;
}

int mpg123_play(struct player *pme)
{
    mpg123_player* me = (mpg123_player*) pme;
    int in_play_thread = pthread_equal(pthread_self(), me->play_thr);

    DBGMSG("mpg123_play(%p) called from thr %ld.\n",
                me, (long)getpid());
    /* check for trivial situations */
    if(me->pl.status == st_play) {
        return 1;
    }
    if(me->pl.status == st_pause) {
        return mpg123_pause(pme);
    }

    if(me->play_thr == (pthread_t)0 || me->play_pid == 0)
        if(!start_play_thread(me)) {
            return 0;
        }

    if(!me->filename) {
        return 0;
    }
    DBGMSG("sending 'LOAD %s'\n", me->filename);
    me->error = 0;
    me->pl.status = st_stop; /* so we can see when the file has been loaded */
    me->wanted_status = st_play;
    cmd_printf(me, "LOAD %s\n", me->filename); 
    if(!in_play_thread)
        WAIT_COND_UNTIL(me->curr_status==st_play || me->error, me->play);
    if(me->pl.status != st_play) {
        me->pl.status = st_play;
        if(me->callback)
            me->callback(&me->pl, CB_STATUS, me->cb_param, NULL);
    }
    if(me->callback)
        me->callback(&me->pl, CB_UPDATE, me->cb_param, NULL);
    return !me->error;
}

/* N.B. synchronous playing is implemented to enable PROFILING (because
   it has problems with threads) and nothing else */
int mpg123_play_sync(struct player *pme)
{
    mpg123_player* me = (mpg123_player*) pme;
    DBGMSG("mpg123_play_sync(%p) called.\n", me);

    if(!me->filename)
        return 0;
    if(me->play_thr != (pthread_t)0)
        terminate_play_thread(me);

    /* uses mpg123/mpg321 "remote control" interface (-R) */
    if(!open_double_pipe(me))
        return 0;

    me->play_thr = pthread_self();
    me->sync = 1;
    DBGMSG("sending 'LOAD %s'\n", me->filename);
    me->error = 0;
    me->pl.status = st_stop;
    me->wanted_status = st_play;
    cmd_printf(me, "LOAD %s\n", me->filename);

    play_thread(me); /* run player synchronously */

    me->play_thr = (pthread_t)0;
    me->sync = 0;
    return me->error;
}

/* 'pause' will pause if playing and resume if already paused
   (and do nothing if stopped) */
int mpg123_pause(struct player *pme)
{
    mpg123_player* me = (mpg123_player*) pme;
    int in_play_thread = pthread_equal(pthread_self(), me->play_thr);

    DBGMSG("mpg123_pause(%p) called from thr %ld.\n",
                me, (long)getpid());
    if(me->play_pid > 0) {
        int st = -1;
        if(me->pl.status == st_play)
            st = st_pause;
        else if(me->pl.status == st_pause)
            st = st_play;
        if(st >= 0) {
            DBGMSG("sending 'PAUSE'\n");
            me->wanted_status = st;
            cmd_printf(me, "PAUSE\n");
            if(!in_play_thread)
                WAIT_COND_UNTIL(me->curr_status == st || me->error, me->play);
            me->pl.status = st;
            if(me->callback)
                me->callback(&me->pl, CB_STATUS, me->cb_param, NULL);
        }
    }
    return 1; /* no error checking implemented */
}

int mpg123_stop(struct player *pme)
{
    mpg123_player* me = (mpg123_player*) pme;
    int in_play_thread = pthread_equal(pthread_self(), me->play_thr);

    DBGMSG("mpg123_stop(%p) called from thr %ld.\n",
                me, (long)getpid());
    if(me->play_pid > 0) {
        if(me->pl.status != st_stop) {
            me->wanted_status = st_stop;
            DBGMSG("sending 'STOP'\n");
            cmd_printf(me, "STOP\n");
            if(!in_play_thread)
                WAIT_COND_UNTIL(me->curr_status == st_stop || me->error, me->play);
            /* comment out to keep mpg123 up all the time: */
            terminate_play_thread(me);
            me->pl.status = st_stop;
            if(me->callback)
                me->callback(&me->pl, CB_STATUS, me->cb_param, NULL);
        }
    }
    return 1;
}

/* jumps 'f' frames from the beginning */
int mpg123_jump_frames(struct player *pme, long f)
{
    mpg123_player* me = (mpg123_player*) pme;
    int in_play_thread = pthread_equal(pthread_self(), me->play_thr);

    DBGMSG("mpg123_jump_frames(%p, %ld) called from thr %ld.\n",
                me, f, (long)getpid());
    /* must be playing for the JUMP command to work */
    if(me->curr_status != st_play) {
        mpg123_play(pme);
        /* mpg123 <=0.59r (not mpg321) is a little buggy and needs this */
        usleep(1000);
    }
    if(!in_play_thread)
        WAIT_COND_UNTIL(me->song_len.frames >= 0, me->play);
    DBGMSG("JUMPing to frame %ld/%ld.\n", f, me->song_len.frames);
    cmd_printf(me, "JUMP %ld\n", f);
    
    return 1; /* still no error checking */
}

/* jumps 'sec' seconds from the beginning */
int mpg123_jump_time(struct player *pme, long sec)
{
    mpg123_player* me = (mpg123_player*) pme;
    int in_play_thread = pthread_equal(pthread_self(), me->play_thr);
    int frame;

    DBGMSG("mpg123_jump_frames(%p, %ld) called from thr %ld.\n",
                me, sec, (long)getpid());
    if(me->song_len.frames < 0) {
        mpg123_play(pme);
        if(!in_play_thread)
            WAIT_COND_UNTIL(me->song_len.frames >= 0, me->play);
    }
    frame = (sec * me->song_len.frames) / me->song_len.seconds;
    return mpg123_jump_frames(pme, frame);
}

/* jumps to `position/total' of the total file (useful for progress bar) */
int mpg123_jump(struct player *pme, long position, long total)
{
    mpg123_player* me = (mpg123_player*) pme;
    int in_play_thread = pthread_equal(pthread_self(), me->play_thr);
    int frame;
    DBGMSG("mpg123_jump(%p, %ld, %ld) called from thr %ld.\n",
                me, position, total, (long)getpid());
    if(me->song_len.frames < 0) {
        mpg123_play(pme);
        if(!in_play_thread)
            WAIT_COND_UNTIL(me->song_len.frames >= 0, me->play);
    }
    frame = (position * me->song_len.frames) / total;
    return mpg123_jump_frames(pme, frame);
}

void get_song_pos_length(mpg123_player* me)
{
    int in_play_thread = pthread_equal(pthread_self(), me->play_thr);
    DBGMSG("get_song_pos_length(%p)\n", me);
    if(me->song_len.frames < 0) {
        if(me->pl.status == st_pause) return; /* shouldn't happen :-/ */
        if(me->pl.status == st_stop) {
            if(me->play_thr == (pthread_t)0 || me->play_pid == 0)
                if(!start_play_thread(me))
                    return;
            DBGMSG("sending 'LOAD %s'\n", me->filename);
            me->wanted_status = st_play;
            cmd_printf(me, "LOAD %s\n", me->filename);
            if(!in_play_thread)
                WAIT_COND_UNTIL(me->curr_status==st_play || me->error, me->play);
        }
        me->error = 0;
        if(!in_play_thread)
            WAIT_COND_UNTIL((me->song_len.frames >= 0 && me->curr_pos.frames >= 0)
                            || me->curr_status != st_play /* song with no frames? */
                            || me->error,
                            me->play);
        pthread_mutex_lock(&me->curr_pos_mtx);
        DBGMSG("got length: (%df, %fs), pos: (%df, %fs)\n",
               me->song_len.frames, me->song_len.seconds,
               me->curr_pos.frames, me->curr_pos.seconds);
        pthread_mutex_unlock(&me->curr_pos_mtx);
        if(me->pl.status == st_stop) {
            DBGMSG("sending STOP\n");
            me->wanted_status = st_stop;
            cmd_printf(me, "STOP\n");
            if(!in_play_thread)
                WAIT_COND_UNTIL(me->curr_status == st_stop || me->error, me->play);
        }           
    }
}

/* query total length (total number of frames/seconds) */
int mpg123_get_length(struct player *pme, song_pos* len)
{   
    mpg123_player* me = (mpg123_player*) pme;
/*  DBGMSG("mpg123_get_length(%p, %p)\n", pme, len); */
    pthread_mutex_lock(&me->curr_pos_mtx);
    if(me->song_len.frames < 0) {
        pthread_mutex_unlock(&me->curr_pos_mtx);
        get_song_pos_length(me);
        pthread_mutex_lock(&me->curr_pos_mtx);
    }
    *len = me->song_len;
    pthread_mutex_unlock(&me->curr_pos_mtx);
    return (len->frames >= 0);
}

/* query current position */
int mpg123_get_pos(struct player *pme, song_pos* pos)
{
    mpg123_player* me = (mpg123_player*) pme;
/*  DBGMSG("mpg123_get_pos(%p, %p)\n", pme, pos); */
    pthread_mutex_lock(&me->curr_pos_mtx);
    if(me->curr_pos.frames < 0) {
        pthread_mutex_unlock(&me->curr_pos_mtx);
        get_song_pos_length(me);
        pthread_mutex_lock(&me->curr_pos_mtx);
    }
    *pos = me->curr_pos;
    pthread_mutex_unlock(&me->curr_pos_mtx);
    return (pos->frames >= 0);
}

/* set callback */
void mpg123_set_callback(struct player *pme, player_cb_func cb, void* param)
{
    mpg123_player* me = (mpg123_player*) pme;
    me->callback = cb;
    me->cb_param = param;
}
