/*
 * wmdimp
 * Copyright (c) 2004 Daniele Tamino <aslo@wmdimp.cjb.net>.
 * This software covered by the GPL.  See COPYING file for details.
 *
 * output.c
 *
 * output module, used by player_mad.c (not by mpg123_player.c) to
 * send audio data to an OSS device (or to a file).
 *
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <pthread.h>
#include <sys/stat.h>

#if !defined(PROFILE)
#  define REGPARM __attribute__((regparm(3)))
#else
#  define REGPARM
#endif

/* by now, use oss only */

#if defined(HAVE_SOUNDCARD_H)
#  include <soundcard.h>
#elif defined(HAVE_SYS_SOUNDCARD_H)
#  include <sys/soundcard.h>
#elif defined(HAVE_MACHINE_SOUNDCARD_H)
#  include <machine/soundcard.h>
#else
#  error "need <soundcard.h>, <sys/soundcard.h> or <machine/soundcard.h>"
#endif

/* needed for the table below */
#ifdef WORDS_BIGENDIAN
#  define BIGENDIAN 1
#else
#  define BIGENDIAN 0
#endif

#if defined(AFMT_S32_LE) && defined(AFMT_S32_BE)
#  define HAVE_AFMT_32BITS
#  ifndef AFMT_S32_NE
#    if WORDS_BIGENDIAN
#      define AFMT_S32_NE AFMT_S32_BE
#    else
#      define AFMT_S32_NE AFMT_S32_LE
#    endif
#  endif
#endif

#include "utils.h"
#include "output.h"
/*#define DBG_MESSAGES 0*/
#include "debug.h"

type_desc_spec output_specs = {
    "oss output",
    "oss_device = +string:'/dev/dsp';"
    "oss_rate = integer:0;"
    "oss_channels = integer:0;"
    "oss_bits = integer:0;"
    "oss_bigendian = boolean:false;" /* big endian input? (mainly for testing) */
    "oss_unsigned = boolean:false;"  /* unsigned input? (mainly for testing) */
    "oss_bufsizems = integer:250;" /* output buffer size in millseconds */
    "oss_bufwait = integer:10;" /* space to wait when buffer is full (ms) */
    "oss_dithering = boolean:true;"
    "oss_verbose = boolean:false;" /* used maily for debugging */
};
 
static int sound_fd = -1; /* file descriptor for /dev/dsp or what */
static str_list* devices; /* list of possible devices */
static char* device = NULL;  /* current device */
#define OUTPUT_TO_DISK 1
static int is_oss = 1; /* if =0 sound_fd refers to a regular disk file */

/* input format (must convert from this) */
static output_fmt in_fmt;
static int in_bps; /* bytes per sample (N.B. including all channels) */
static int in_bpc = -1; /* bytes per channel (=in_bps / in_fmt.chan) */

/* format specified (forced) in config file for output */
static output_fmt force_fmt;
static int verbose = IFDBG(1, 0); /* print various information messages? */
static int cfg_dither = 1; /* 0 = disable dithering */

/* output format */
static output_fmt out_fmt; /* actual format */
static int out_bps = -1; /* bytes per sample */
static int out_bpc = -1; /* bytes per channel (=out_bps / out_fmt.chan) */

static int bitsdiff = 0; /* = abs(in_fmt.bits - out_fmt.bits) */ 
static int samp_max = 0;

static int oss_fmt = -1; /* one of AFMT_??? */
static int oss_swapin = 0; /* 1 = swap high-low bytes or words in input */
static int oss_unsig_in = 0; /* 1 = conv. signed to unsigned */
/* N.B. the *in conversion are mainly for testing, because output_write
   will overwrite input data, which is a Bad Thing */
static int oss_swapout = 0; /* 1 = swap high-low bytes or words in output */
static int oss_unsig_out = 0; /* 1 = conv. signed to unsigned */

static unsigned int cfg_bufsizems = 250; /* defaults to 250 msec */
static unsigned int cfg_bufminspace = 10;
static unsigned int oss_fragsize = -1; /* oss fragment size (GETBLKSIZE) */
static unsigned int oss_nfrags = -1;   /* oss n of fragments (GETOSPACE) */
static unsigned int oss_minspace = 0; /* space to wait in buffers before write */
static int select_works = 1;
typedef struct buffer {
    char* start; /* first byte of buffer */
    char* end;   /* first byte past the end (size = end - start) */
    char* head;  /* first free byte */
    char* tail;  /* first written byte */
    int full; /* when head==tail, it is empty of full? read this! */
} buffer;
static buffer tmpbuf;

static inline int buffer_size(buffer* buf) { return buf->end - buf->start; }

static inline int buffer_data(buffer* buf)
{
    if(buf->tail < buf->head) {
        return buf->head - buf->tail;
    } else if(buf->tail > buf->head)
        return (buf->end - buf->start) - (buf->tail - buf->head);
    else if(buf->full)
        return buf->end - buf->start;
    return 0;
}

static inline int buffer_space(buffer* buf)
{
    if(buf->head < buf->tail) {
        return buf->tail - buf->head;
    } else if(buf->head > buf->tail)
        return (buf->end - buf->start) - (buf->head - buf->tail);
    else if(!buf->full)
        return buf->end - buf->start;
    return 0;
}

/* returns the max number of contiguous bytes that can be written in
   the address returned in p (if != NULL) */
static inline int buffer_write(buffer* buf, char** p)
{
    int i;
    if(buf->head >= buf->tail) {
        if(buf->head == buf->tail && buf->full)
            i = 0;
        else
            i = buf->end - buf->head;
    } else
        i = buf->tail - buf->head;
    if(p)
        *p = buf->head;
    return i;
}

/* returns the max number of contiguous bytes that can be read from
   the address returned in p (if != NULL) */
static inline int buffer_read(buffer* buf, char** p)
{
    int i;
    if(buf->tail >= buf->head) {
        if(buf->tail == buf->head && !buf->full)
            i = 0;
        else
            i = buf->end - buf->tail;
    } else
        i = buf->head - buf->tail;
    if(p)
        *p = buf->tail;
    return i;
}

/* advance the head n bytes */
static inline void buffer_put(buffer* buf, int n)
{
    assert(n <= buf->end - buf->start);
    /*     DBGMSG("buffer_put: %d bytes\n", n); */
    if(n > 0) {
        buf->head += n;
        if(buf->head >= buf->end)
            buf->head -= buf->end - buf->start;
        if(buf->head == buf->tail) buf->full = 1;
    }
}

/* advance the tail n bytes */
static inline void buffer_get(buffer* buf, int n)
{
    assert(n <= buf->end - buf->start);
/*     DBGMSG("buffer_get: %d bytes\n", n); */
    if(n > 0) {
        buf->tail += n;
        if(buf->tail >= buf->end)
            buf->tail -= buf->end - buf->start;
        if(buf->tail == buf->head)
            buf->tail = buf->head = buf->start;
        buf->full = 0;
    }
}

static inline int buffer_full(buffer* buf) { return buf->full; }
static inline int buffer_empty(buffer* buf)
  { return (buf->head == buf->tail) && !buf->full; }

/* signalled when some data is put in or read from the buffer */
pthread_cond_t buf_cond = PTHREAD_COND_INITIALIZER;
/* mutex associated with buf_cond */
pthread_mutex_t buf_mtx = PTHREAD_MUTEX_INITIALIZER;
static enum {
    OMSG_RUN,
    OMSG_TERM,
    OMSG_TERM_NOW, /* do not flush buffer */
    OMSG_PAUSE,
    OMSG_FLUSH
} out_thr_msg;
int term_output_thread(int flush); /* if(flush) flush buffer before exit */
int start_output_thread(void);

enum conv_types {
    conv_none,     /* input format bitwise equal to output format */
    conv_copy,     /* same rate, different sample format */
    conv_resample  /* different rate */
};
static int conv_type = conv_none;

typedef struct fmt_table_entry {
    int fmt; /* AFMT_??? */
    const char *desc; /* description */
    int supp;  /* is it supported ? */
    int bits;  /* bits per sample  (bytes per sample = (bits + 7)/8 ) */
    int swap;  /* need to swap bytes */
    int unsig; /* need to convert from signed to unsigned */
} fmt_table_entry;

fmt_table_entry fmt_tab[] = {
    { AFMT_MU_LAW, "mu-law", 0, 8, 0, 0 }, /* unsupported */
    { AFMT_A_LAW, "a-law", 0, 8, 0, 0 }, /* unsupported */
    { AFMT_IMA_ADPCM, "IMA ADPCM", 0, 4, 0, 0 }, /* unsupported */
    { AFMT_U8, "8 bit unsigned", 1, 8, 0, 1 },
    { AFMT_S8, "8 bit signed", 1, 8, 0, 0 },
    { AFMT_U16_LE, "16 bit unsigned little-endian", 1, 16, BIGENDIAN, 1 },
    { AFMT_S16_LE, "16 bit signed little-endian", 1, 16, BIGENDIAN, 0 },
    { AFMT_U16_BE, "16 bit unsigned big-endian", 1, 16, !BIGENDIAN, 1 },
    { AFMT_S16_BE, "16 bit signed big-endian", 1, 16, !BIGENDIAN, 0 },
#ifdef HAVE_AFMT_32BITS
    { AFMT_S32_LE, "24/32 bit little-endian", 1, 16, BIGENDIAN, 0 },
    { AFMT_S32_BE, "24/32 bit big-endian", 1, 16, !BIGENDIAN, 0 },
#endif
    { AFMT_MPEG, "MPEG", 0, -1, 0, 0 }, /* unsupported  */
    { -1, "unknown", 0, 0, 0, 0 } /* terminator */
};

fmt_table_entry* get_fmt_info(int fmt)
{
    fmt_table_entry* e;
    for(e = fmt_tab; e->fmt >= 0; e++)
	if(e->fmt == fmt)
	    return e;
    return e; /* terminator is correctly marked as not supported ;-) */
}

const char* const fmtstr(int fmt)
{
    fmt_table_entry *e = get_fmt_info(fmt);
    if(e) return e->desc;
    else  return "(unknown format)";
}

/* Resampling structures and functions */

/* use Bresenham's Algorithm (for up- and down-sampling) */
typedef struct resample_state {
    int state;
    int max;   /* max rate of the two (in and out) */
    int inc1;  /* increment when state < 0 */
    int inc2;  /* incremente when state >= 0 (==inc1 + 1) */
    int step1; /* add to state when state < 0 */
    int step2; /* add to state when state >= 0 */
} resample_state;

resample_state resamp_state;

typedef struct quantize_state {
    /* some constant precalculated info */
    int bits; /* n. of bits to remove */
    int mask; /* mask of 'bits' bits ((1 << bits) - 1)  */
    int32_t min, max; /* max and min sample values (before shifting) */
    int avoid_ovfl; /* avoid overflow (lose accuracy) when input is 32 bit */
    /* dither state */
    int32_t state;
#ifndef ERROR_DIFFUSION
    int32_t seed;
#endif
    int32_t err[3]; /* 3rd order noise shaping */
    int32_t nsadd; /* precalc. noise shaping bias */
    int32_t bias; /* = 1 << (bits-1) */
} quantize_state;

void init_quantizer(quantize_state* st, int ibits, int obits);
inline int32_t quantize(int32_t samp, quantize_state* st)
     REGPARM;

typedef int32_t (*read_sample_func)(void* data)
     REGPARM;
static read_sample_func read_sample = NULL;

typedef void (*write_sample_func)(void* data, int32_t value)
     REGPARM;
static write_sample_func write_sample = NULL;

/* used to read the output and write the input (this last ONLY FOR TESTING,
   it's not wise to modify input data IMHO) */
static read_sample_func read_sample_out = NULL;
static write_sample_func write_sample_in = NULL;

typedef struct chan_info {
    quantize_state qst; /* no other info by now */
} chan_info;

/* output channels info */
chan_info* channels = NULL;

/* use many ultra-specialized (and ultra fast) small functions
   to convert from input to output format.
   When I say 'many' I mean 162 of those :) */
typedef int (*copy_func)(char* in, char* out, long n);
static copy_func copy_samples = NULL;
typedef int (*resample_func)(char* in, char* out, long nin,
                 long nout, long* ni, resample_state* st);
static resample_func resample_samples = NULL;

/* function to convert from input to output channel samples
   (= resize samples with or without dithering)*/
typedef int32_t (*convert_channel_func)(int32_t samp, chan_info* info) REGPARM;
static convert_channel_func convert_chan = NULL;

/* function for pre- and post-processing (= convert bigendian/unsigned) */
typedef int32_t (*convert_func)(int32_t samp)
     REGPARM;
static convert_func convert_in = NULL; /* pre-processing */
static convert_func convert_out = NULL; /* post-processing */
/* N.B. pre-processing is ONLY FOR TESTING because it writes on the input
   buffer given by the caller, something's (rather) fast but not so clean */

int select_read_write_func(int ibpc, int obpc);
int select_convert_channel_func(void);
int select_pre_post_convert_func(void);
void select_copy_samples_func(int ichn, int ochn, int ibpc, int obpc);
inline int select_functions(int ibpc, int obpc, int ichn, int ochn)
{
    if(!select_read_write_func(ibpc, obpc)) return 0;
    if(!select_convert_channel_func()) return 0;
    select_copy_samples_func(ichn, ochn, ibpc, obpc);
    if(!select_pre_post_convert_func()) return 0;
    return 1;
}

void init_resample(int rate_in, int rate_out, resample_state* st);

/* if needed, allocate temporary buffer */
static int allocate_temp_buffer(void)
{
    unsigned long bufsize;
    oss_fragsize = 0;
    if(tmpbuf.start == NULL) {
        if(is_oss) {
            int ret;
            /* get oss fragment size */
            audio_buf_info binfo;
            ret = ioctl(sound_fd, SNDCTL_DSP_GETOSPACE, &binfo);
            if(ret == 0) {
                if(verbose) {
                    fprintf(stderr, "%s: n. frags = %d, frag size = %d\n",
                            device, binfo.fragstotal, binfo.fragsize);
                }
                oss_fragsize = binfo.fragsize;
                oss_nfrags = binfo.fragstotal;
                oss_minspace = out_bps * out_fmt.rate * cfg_bufminspace / 1000;
            } else {
                oss_fragsize = 8192;
                oss_nfrags = 0;
                oss_minspace = 0;
                DBGMSG("%s: cannot get fragment size, I'll do without\n", device);
            }
        }
        bufsize = (unsigned)((float)cfg_bufsizems * out_bps * out_fmt.rate / 1000);
        bufsize = bufsize / out_bps * out_bps; /* must be multiple of out_bps */
        DBGMSG("%s: bufsizems = %d, calculated bufsize = %u\n",
               device, cfg_bufsizems, bufsize);
        if(bufsize < oss_fragsize)
            bufsize = oss_fragsize;
        if(!oss_minspace)
            oss_minspace = bufsize * 15/100;
        /* allocate buffer */
        tmpbuf.start = malloc(bufsize);
        if(!tmpbuf.start) {
            fprintf(stderr, "output_write: Out of memory allocating"
                    " temp buffer (%ld bytes)\n", bufsize);
            return 0;
        } 
        tmpbuf.end = tmpbuf.start + bufsize;
        tmpbuf.full = 0;

        if(verbose) {
            printf("%s: allocated %ld bytes (= %.0f ms) of temporary space.\n",
                   device, bufsize, bufsize * 1000.0 / (out_fmt.rate * out_bps));
            DBG(printf("%s: minimum space to wait is %d (= %.0f ms)\n",
                       device, oss_minspace,
                       oss_minspace* 1000.0 / (out_fmt.rate * out_bps));)
        }
        tmpbuf.head = tmpbuf.tail = tmpbuf.start;
    }
    return 1;
}

static inline void free_temp_buffer(void)
{
    if(tmpbuf.start) {
        free(tmpbuf.start);
        tmpbuf.start = tmpbuf.end = tmpbuf.head = tmpbuf.tail = NULL;
        tmpbuf.full = 0;
        /*  tmpbuf = NULL; */
        /*  tmp_n = 0; */
    }
}

/* Now, the interface */

/* Load configuration options */


int output_init(opt_hash* ht)
{
    opt_tree* t;
    t = opt_hash_find(ht, "oss_verbose");
    if(t) {
        assert(t->root && t->root->next);
        verbose = t->root->next->val.b;
    }
    t = opt_hash_find(ht, "oss_device");
    if(t) {
        assert(t->root && t->root->next);
        load_str_list(t->root->next, &devices);
    }
    t = opt_hash_find(ht, "oss_rate");
    if(t) {
        assert(t->root && t->root->next);
        force_fmt.rate = t->root->next->val.i;
    }
    t = opt_hash_find(ht, "oss_bits");
    if(t) {
        assert(t->root && t->root->next);
        force_fmt.bits = t->root->next->val.i;
    }
    t = opt_hash_find(ht, "oss_channels");
    if(t) {
        assert(t->root && t->root->next);
        force_fmt.chan = t->root->next->val.i;
    }
    t = opt_hash_find(ht, "oss_bigendian");
    if(t) {
        assert(t->root && t->root->next);
        oss_swapin = t->root->next->val.b;
        DBG(if(oss_swapin && verbose)
            printf("output.c: input is reverse-endian (from config file)\n"));
    }
    t = opt_hash_find(ht, "oss_bufsizems");
    if(t) {
        assert(t->root && t->root->next);
        cfg_bufsizems = t->root->next->val.i;
    }
    t = opt_hash_find(ht, "oss_bufwait");
    if(t) {
        assert(t->root && t->root->next);
        cfg_bufminspace = t->root->next->val.i;
    }
    t = opt_hash_find(ht, "oss_unsigned");
    if(t) {
        assert(t->root && t->root->next);
        oss_unsig_in = t->root->next->val.b;
        DBG(if(oss_unsig_in && verbose)
            printf("output.c: input is unsigned (from config file)\n"));
    }
    t = opt_hash_find(ht, "oss_dithering");
    if(t) {
        assert(t->root && t->root->next);
        cfg_dither = t->root->next->val.b;
    }
    DBG(if(verbose) printf("output.c: dithering %s\n",
                           cfg_dither ? "enabled":"disabled");)

        return 1;
}

void output_finish(void)
{
    if(devices) {
	free_str_list(devices);
	devices = NULL;
    }
}

#ifdef OUTPUT_TO_DISK
int output_open_to_disk(output_fmt* fmt)
{
    if(sound_fd < 0) {
        sound_fd = open(device, O_WRONLY|O_CREAT|O_TRUNC|O_NONBLOCK);
        if(sound_fd >= 0) {
            /* we want normal blocking calls */
            int fl = fcntl(sound_fd, F_GETFL, 0);
            if(fl < 0) { close(sound_fd); return 0; } /* shouldn't happen */
            fcntl(sound_fd, F_SETFL, fl & (~O_NONBLOCK));
        } else {
            fprintf(stderr, "%s: open: %s\n", device, strerror(errno));
            return 0;
        }
    } else {
        ftruncate(sound_fd, 0);
    }
    is_oss = 0; /* mark that we're writing to a regular file */
    oss_swapout = 0;
    oss_unsig_out = 0;
    out_fmt = in_fmt = *fmt;
    if(force_fmt.rate > 0) out_fmt.rate = force_fmt.rate;
    if(force_fmt.bits > 0) out_fmt.bits = force_fmt.bits;
    if(force_fmt.chan > 0) out_fmt.chan = force_fmt.chan;
    in_bpc = (in_fmt.bits + 7) >> 3;
    in_bps = in_bpc * in_fmt.chan;
    out_bpc = (out_fmt.bits + 7) >> 3;
    /* use full range when writing to file */
    out_fmt.bits = out_bpc << 3;
    out_bps = out_bpc * out_fmt.chan;
    printf("%s: writing with %d bits, %d channels @ %ld Hz\n",
           device, out_fmt.bits, out_fmt.chan, out_fmt.rate);
    if(in_fmt.rate != out_fmt.rate)
        conv_type = conv_resample;
    else if(in_fmt.bits != out_fmt.bits ||
            in_fmt.chan != out_fmt.chan)
        conv_type = conv_copy;
    else
        conv_type = conv_none;
    bitsdiff = abs(in_fmt.bits - out_fmt.bits);
    samp_max = (((1 << (out_fmt.bits-1)) - 1) << bitsdiff);
    return 1;
}
#else
int output_open_to_disk() { return 0; }
#endif

/* Opens the preferred audio device, as specified in the config file.
   If format is not supported by the device, it will be emulated
   in software (downsampling, quantization, chan. merging).
   By now, downsampling is done straight (no interpolation/antialias),
   quantization is done with dithering and noise shaping,
   channel merging is done after downsampling and before quantization.
   The actual format is stored in actual_fmt (if != NULL).
   Return value is 0 on error and !=0 on success.
   N.B. By now, fmt->bits cannot be greater than 32, and channel
   merging works only from N (stereo or more) to 1 (mono). If these condition
   are not met, the call fails (returns 0).
 */
#define CHECKERR(ret, msg) \
    if((ret) < 0) {\
        fprintf(stderr, "%s: " msg ": %s\n", device, strerror(errno)); \
        close(sound_fd); \
        sound_fd = -1; \
        return 0; \
    }

int get_best_format(int bits)
{
    /* check if the 'bits' parameter is valid and convert it */
    if(bits <= 8) {
        return AFMT_S8;
#ifdef AFMT_S32_NE
    } else if(bits <= 16) {
        return AFMT_S16_NE;
#endif
    } else if(bits <= 32) {
#ifdef AFMT_S32_NE
        return AFMT_S32_NE;
#else
        return AFMT_S16_NE;
#endif
    } else {
        return -1;
    }
}

int output_open(output_fmt* fmt, output_fmt* actual_fmt)
{
    struct stat st;
    int ret, form, chn, rate;
    int fragsize, frag, caps;
    fmt_table_entry* e = NULL;
    if(sound_fd < 0 || memcmp(fmt, &in_fmt, sizeof(output_fmt)) != 0) {
        int bits, fl;
        if(sound_fd >= 0 && is_oss) {
            term_output_thread(1);
            /* OSS Programmer's Guide advices everywhere to close and
               reopen the device instead of using SND_DSP_RESET... why? */
            close(sound_fd);
            free_temp_buffer();
            if(channels) {
                free(channels);
                channels = NULL;
            }
        }

        if(force_fmt.bits > 0) bits = force_fmt.bits;
        else bits = fmt->bits;
        if(bits > 32) {
            fprintf(stderr, "output_open: unsupported format (bits = %d).\n",
                    bits);
            return 0;
        }
    
        if(device)
            sound_fd = open(device, O_WRONLY | O_NONBLOCK);
        if(sound_fd < 0) { /* no device selected or device unavailable */
            str_list *l;
            for(l = devices; l != NULL; l = l->next) {
                device = l->str;
                if((sound_fd = open(device, O_WRONLY | O_NONBLOCK)) >= 0)
                    break;
            }
            /* still no available device found? */
            if(sound_fd < 0 && !output_open_to_disk(fmt)) {
                fprintf(stderr, "%s: open: %s\n", device, strerror(errno));
                return 0;
            }
        }
        /* we want normal blocking calls */
        fl = fcntl(sound_fd, F_GETFL, 0);
        if(fl < 0) { close(sound_fd); return 0; } /* shouldn't happen */
        fcntl(sound_fd, F_SETFL, fl & (~O_NONBLOCK));

        /* try to see if this is a OSS DSP device */    
        if((!fstat(sound_fd, &st) && !S_ISCHR(st.st_mode)) ||
           ((ret=ioctl(sound_fd, SNDCTL_DSP_GETCAPS, &caps)) < 0 &&
            errno == 25 /*=ENOTTY why? boh...*/)) {
            ret = output_open_to_disk(fmt) - 1;
            if(ret < 0) {
                fprintf(stderr, "%s: open: %s\n", device, strerror(errno));
                return 0;
            }
            goto setup_structs;
        }

        in_fmt = *fmt;
        out_fmt.rate = fmt->rate;
        out_fmt.chan = fmt->chan;
    
        /* round up to 8, 16 or 32 bits (24 bit not supported) */
        bits = (bits + 7) & (~7);
        if(bits == 24) bits = 32;

        /* set the fragment size. oss suggest to set it as early as possible,
           but now we don't know the output format, so use input.
           Later we'll try to set it again (at worst it will fail). */
        fragsize = (fmt->rate * ((fmt->bits+7) / 8) * fmt->chan)
            * cfg_bufsizems / (1000 * 2); /* 1/2 of buffer size */
        for(frag = 1; (1 << frag) < fragsize; frag++);
        DBGMSG("%s: trying to set fragment size to %d (2^%d)\n",
               device, 1<<frag, frag);
        frag |= 32 << 16;
        ioctl(sound_fd, SNDCTL_DSP_SETFRAGMENT, &frag);

        /* try to set desired output format */
        for(; bits >= 8; bits /= 2) {
            form = oss_fmt = get_best_format(bits);
            ret = ioctl(sound_fd, SNDCTL_DSP_SETFMT, &oss_fmt);
            CHECKERR(ret, "setting output format");
            e = get_fmt_info(oss_fmt);
            if(e->supp) {
                if(form == oss_fmt || e->bits >= bits) {        
                    bits = e->bits;
                    out_bps = out_bpc = (bits + 7) / 8;
                    oss_swapout = e->swap;
                    oss_unsig_out = e->unsig;
                    if(bits == fmt->bits)
                        conv_type = conv_none;
                    else
                        conv_type = conv_copy;
                    break;
                }
            }
            /* else continue and try reducing bits */
        }
        if(bits < 8) {
            fprintf(stderr, "output_open: %s: cannot set sample format (bits=%d)\n",
                    device, fmt->bits);
            close(sound_fd);
            sound_fd = -1;
            return 0;
        }
        out_fmt.bits = bits;
        bitsdiff = e->bits > fmt->bits ? e->bits - fmt->bits
            : fmt->bits - e->bits;
        samp_max = (((1 << (e->bits-1)) - 1) << bitsdiff);
        if(verbose)
            printf("%s: sample format: %s\n", device, e->desc);

        /* set channels */
        if(force_fmt.chan > 0) chn = force_fmt.chan;
        else chn = fmt->chan;
        ret = ioctl(sound_fd, SNDCTL_DSP_CHANNELS, &chn);
        CHECKERR(ret, "setting channels");
        if(chn != fmt->chan) {
            out_fmt.chan = chn;
            conv_type = conv_copy;
        }
        if(verbose)
            printf("%s: %d output channels\n", device, chn);
        out_bps *= chn;

        /* set rate */
        if(force_fmt.rate > 0) rate = force_fmt.rate;
        else rate = fmt->rate;
        ret = ioctl(sound_fd, SNDCTL_DSP_SPEED, &rate);
        CHECKERR(ret, "setting rate");
        if(rate != fmt->rate) {
            out_fmt.rate = rate; /* need resampling */
            conv_type = conv_resample;      
        }
        if(verbose)
            printf("%s: output rate: %d Hz\n", device, rate);

        if(out_fmt.rate != in_fmt.rate || out_bps != in_bps) {
            fragsize = (out_fmt.rate * out_bps * out_fmt.chan)
                * cfg_bufsizems / (1000 * 2); /* 1/2 of buffer size */
            for(frag = 1; (1 << frag) < fragsize; frag++);
            DBGMSG("%s: trying to re-set fragment size to %d (2^%d)\n",
                   device, 1<<frag, frag);
            frag |= 32 << 16;
            /* this may fail, but we don't care */
            ioctl(sound_fd, SNDCTL_DSP_SETFRAGMENT, &frag);
        }

    setup_structs:
        /* initialize other structures */
        in_bps = (in_bpc = ((fmt->bits + 7) / 8) ) * fmt->chan;
        /* select read_sample, write_sample and convert */
        if(!select_functions(in_bpc, out_bpc, fmt->chan, out_fmt.chan)) {
            close(sound_fd);
            sound_fd = -1;
            return 0;
        }
        if(!allocate_temp_buffer()) {
            close(sound_fd);
            sound_fd = -1;
            return 0;
        }
        if(conv_type != conv_none) {
            if(conv_type == conv_resample)
                init_resample(in_fmt.rate, out_fmt.rate, &resamp_state);
            if(out_fmt.bits < fmt->bits && cfg_dither) {
                int i;
                channels = (chan_info*)malloc(out_fmt.chan*sizeof(chan_info));
                if(!channels) {
                    fprintf(stderr, "BuyALotOfRam NOW! (= error: out of memory)\n");
                    close(sound_fd);
                    sound_fd = -1;
                    free_temp_buffer();
                    return 0;
                }
                for(i = 0; i < out_fmt.chan; i++)
                    init_quantizer(&channels[i].qst, fmt->bits, out_fmt.bits);
            }
        }
        if((ret=start_output_thread()) != 0) {
            fprintf(stderr, "output_open: pthread_create: %s\n",
                    strerror(ret));
            close(sound_fd);
            sound_fd = -1;
            free_temp_buffer();
            return 0;
        }
    } 
    if(actual_fmt)
        *actual_fmt = out_fmt;
    return 1;
}


/* Close audio device; if was playing then stop as with output_stop(0, 0).
 */
int output_close(void)
{
    int ret = 0;
    term_output_thread(1);
    if(sound_fd >= 0) {
        if(is_oss)
            ioctl(sound_fd, SNDCTL_DSP_POST, 0); /* flush all data */
        ret = close(sound_fd);
        if(ret == 0)
            sound_fd = -1;
        free_temp_buffer();
    }
    if(channels) {
        free(channels);
        channels = NULL;
    }
    return ret == 0;
}

/* stops playing immediately; do not discard audio data not yet processed.
   when called a second time, resume from where it was stopped.
   If this is not supported, this call is equivalent to output_stop(0, 1)
   or, if that isn't supported too, to output_stop(1, 0).
   This call will never block.
*/
int output_pause(void)
{
    out_thr_msg = OMSG_PAUSE;
    return 1;
}


/* stops playing; if now == 1 stop immediately and discard unprocessed
   data, else wait for all data to be processed, then stop.
   If now == 0, data is finished processing asynchronously if async == 1,
   else the program is blocked until all data has been processed.
   >>> Maybe async isn't very useful, so I may remove it <<<
*/
int output_stop(int now, int async)
{
    int ret = 0;
    if(now) {
        pthread_mutex_lock(&buf_mtx);
        tmpbuf.head = tmpbuf.tail = tmpbuf.start; /* empty buffer */
        tmpbuf.full = 0;
        DBGMSG("output_stop: signalling buf_cond...\n");
        pthread_cond_broadcast(&buf_cond);
        ret = ioctl(sound_fd, SNDCTL_DSP_RESET, 0);
        pthread_mutex_unlock(&buf_mtx);
    } else {
        out_thr_msg = OMSG_FLUSH;
        if(!async) {
            pthread_mutex_lock(&buf_mtx);
            while(!buffer_empty(&tmpbuf)) {
                DBGMSG("output_stop: buffer: %db, waiting for empty buffer\n",
                       buffer_data(&tmpbuf));
                pthread_cond_wait(&buf_cond, &buf_mtx);
            }
            pthread_mutex_unlock(&buf_mtx);
            ret = ioctl(sound_fd, SNDCTL_DSP_SYNC, 0);
        }
    }
    return (ret == 0);
}

/* and now, the big part: sending data to the device, making all needed
   conversions (resampling, quantizing, channel processing (stereo-->mono) ).
   All conversions are done internally with 32 bit of precision.
*/

void init_resample(int rate_in, int rate_out, resample_state* st)
{
    st->max = rate_in > rate_out ? rate_in : rate_out;
    st->inc1 = rate_in / rate_out;
    st->inc2 = st->inc1 + 1;
    st->step1 = 2 * (rate_in % rate_out);
    st->step2 = st->step1 - 2 * rate_out;
    st->state = st->step1 - rate_out;
}

void init_quantizer(quantize_state* st, int ibits, int obits)
{
    if(ibits == 32) {
        /* if ibits == 32 some variables may overflow while quantizing,
           so, lose 1 LSBit to avoid overflow when ibits == 32 */
        st->bits = (ibits - obits - 1);
        st->avoid_ovfl = 1;
    } else {
        st->bits = (ibits - obits);
        st->avoid_ovfl = 0;
    }
    st->mask = (1 << st->bits) - 1;
    st->max = (((1 << (obits-1)) - 1) << st->bits);
    st->min = -st->max - 1;

    st->state = 0;
#ifndef ERROR_DIFFUSION
    st->seed = time(0);
#endif
    st->err[0] = st->err[1] = st->err[2] = 0;
    st->nsadd = 0;
    st->bias = (1 << (st->bits-1));
}

#ifndef ERROR_DIFFUSION /* error diffusion doesn't need a random generator  */
/* pseudo-random number generator  */
static inline int32_t rnd(int32_t s) {
    return (s * 145987621 + 1) /*& 0xFFFFFFFFL */;
}
#endif

/* perform quantization with dithering and noise shaping */
inline int32_t quantize(int32_t x, quantize_state* s)
{
    int32_t u, v, aux;
    int bits = s->bits; /* load into registers */

    if(s->avoid_ovfl)
        x >>= 1; /* lose 1 low bit and avoid overflows */

    /* noise shaping */
    u = x + s->nsadd;

#ifndef ERROR_DIFFUSION
    /* random dither */
    s->seed = rnd(s->seed);
    aux = s->seed >> (32-bits); /* take high (more random) bits */
    v = u + (aux - s->state);
    s->state = aux;
#else
    /* error diffusion dither - guess if it's better for audio as it
       is for graphics ... it sounds better (IMHO), but only with noise shaping
       Surely, it's faster :-)
    */
    (void) aux;
    v = u - s->state;
    s->state += (v & ~s->mask) - u;
#endif

    /* clip */
    if(v > s->max)      v = s->max;
    else if(v < s->min) v = s->min;

    s->err[2] = s->err[1];
    s->err[1] = s->err[0]/2;
    s->err[0] = (v & ~s->mask) - u;
    /* precalculate noise shaping for the next sample
       (faster because we have all values loaded into registers) */
    s->nsadd = -s->err[0] + s->err[1] - s->err[2] + s->bias;

    return v >> bits;
}

#define DECLARE_READ_SAMPLE(name, bits) \
int32_t read_sample_##name(void* data) REGPARM;\
inline int32_t read_sample_##name(void* data) \
{ \
    return *(int##bits##_t *) data; \
}

DECLARE_READ_SAMPLE(1, 8);
DECLARE_READ_SAMPLE(2, 16);
DECLARE_READ_SAMPLE(4, 32);

#define DECLARE_WRITE_SAMPLE(name, bits) \
void write_sample_##name(void* data, int32_t value) REGPARM;\
inline void write_sample_##name(void* data, int32_t value) \
{ \
    (*(int##bits##_t *) data) = (int##bits##_t) value; \
}

DECLARE_WRITE_SAMPLE(1, 8);
DECLARE_WRITE_SAMPLE(2, 16);
DECLARE_WRITE_SAMPLE(4, 32);

int select_read_write_func(int ibpc, int obpc)
{
    switch(ibpc) {
    case 1: read_sample = read_sample_1;
        write_sample_in = write_sample_1;
        break;
    case 2: read_sample = read_sample_2;
        write_sample_in = write_sample_2;
        break;
    case 4: read_sample = read_sample_4;
        write_sample_in = write_sample_4;
        break;
    default:
        DBGMSG("BUGBUG: select_read_write_func(%d, %d): invalid ibpc\n",
               ibpc, obpc);
        read_sample = NULL;
        write_sample_in = NULL;
        break;
    }
    switch(obpc) {
    case 1: write_sample = write_sample_1;
        read_sample_out = read_sample_1;
        break;
    case 2: write_sample = write_sample_2;
        read_sample_out = read_sample_2;
        break;
    case 4: write_sample = write_sample_4;
        read_sample_out = read_sample_4;
        break;
    default:
        DBGMSG("BUGBUG: select_read_write_func(%d, %d): invalid obpc\n",
               ibpc, obpc);
        DBG(abort());
        write_sample = NULL;
        read_sample_out = NULL;
        break;
    }
    return (read_sample != NULL && write_sample != NULL);
}

/* swapping bytes and words (little --> big endian and vice versa) */
#define DO_SWAP_8(samp) (samp)
#define DO_SWAP_16(samp) \
  ((unsigned short)samp >> 8) | (((unsigned short)(samp) & 0xFF) << 8) 
#define DO_SWAP_32(samp) \
  ((DO_SWAP_16((unsigned long)(samp) >> 16)) | \
   DO_SWAP_16(((unsigned long)(samp) & 0xFFFF) << 16))

/* convert signed to unsigned */
#define DO_UNSIG_8(samp) ((samp) ^ 0x80)
#define DO_UNSIG_16(samp) ((samp) ^ 0x8000)
#define DO_UNSIG_32(samp) ((samp) ^ 0x80000000)

#define DECLARE_CONVERT_SWAP(bits) \
int32_t convert_swap_##bits(int32_t samp) REGPARM; \
int32_t convert_swap_##bits(int32_t samp) { \
    return DO_SWAP_##bits(samp); \
}

#define DECLARE_CONVERT_UNSIG(bits) \
int32_t convert_unsig_##bits(int32_t samp) REGPARM; \
int32_t convert_unsig_##bits(int32_t samp) { \
    return DO_UNSIG_##bits(samp); \
}

#define DECLARE_CONVERT_SWAP_UNSIG(bits) \
int32_t convert_swap_unsig_##bits(int32_t samp) REGPARM; \
int32_t convert_swap_unsig_##bits(int32_t samp) { \
    int32_t x = DO_SWAP_##bits(samp); \
    return DO_UNSIG_##bits(x); \
}

#define DECLARE_CONVERT_UNSIG_SWAP(bits) \
int32_t convert_unsig_swap_##bits(int32_t samp) REGPARM;\
int32_t convert_unsig_swap_##bits(int32_t samp) { \
    int32_t x = DO_UNSIG_##bits(samp); \
    return DO_SWAP_##bits(x); \
}

DECLARE_CONVERT_SWAP(8);
DECLARE_CONVERT_SWAP(16);
DECLARE_CONVERT_SWAP(32);
DECLARE_CONVERT_UNSIG(8);
DECLARE_CONVERT_UNSIG(16);
DECLARE_CONVERT_UNSIG(32);
DECLARE_CONVERT_SWAP_UNSIG(8); /* for input */
DECLARE_CONVERT_SWAP_UNSIG(16);
DECLARE_CONVERT_SWAP_UNSIG(32);
DECLARE_CONVERT_UNSIG_SWAP(8); /* for output */
DECLARE_CONVERT_UNSIG_SWAP(16);
DECLARE_CONVERT_UNSIG_SWAP(32);

/* quantizing/resizing samples functions */

int32_t convert_dither(int32_t samp, chan_info* chi) REGPARM;
int32_t convert_dither(int32_t samp, chan_info* chi)
{
    return quantize(samp, &chi->qst);
}

/* round and scale */
int32_t convert_resize_down(int32_t samp, chan_info* chi) REGPARM;
int32_t convert_resize_down(int32_t samp, chan_info* chi)
{
    int32_t x = samp + (1L << (bitsdiff-1));
    (void) chi; /* avoid warnings */
    if(x > samp_max) x = samp_max;
    else if(x < -samp_max - 1) x = -samp_max - 1;
    return x >> bitsdiff;
}

int32_t convert_resize_up(int32_t samp, chan_info* chi) REGPARM;
int32_t convert_resize_up(int32_t samp, chan_info* chi)
{
    (void) chi;
    if(samp > samp_max) samp = samp_max;
    else if(samp < -samp_max - 1) samp = -samp_max - 1;
    return (samp << bitsdiff);
}

int32_t convert_noconv(int32_t samp, chan_info* chi) REGPARM;
int32_t convert_noconv(int32_t samp, chan_info* chi)
{ (void) chi; return samp; }

int select_convert_channel_func(void)
{
    if(in_fmt.bits == out_fmt.bits) {
        convert_chan = convert_noconv;
    } else if(in_fmt.bits < out_fmt.bits) {
        convert_chan = convert_resize_up; 
    } else {
        if(cfg_dither)
            convert_chan = convert_dither;
        else
            convert_chan = convert_resize_down;
    }
    return 1;
}

/* converting from all to all sample format: each conversion has its own
   small and fast convert function: about a hundred of those funcs! */

static char** input_data = NULL;

/* [ichn][ochn][ibpc-1][obpc-1]
   ichn and ochn are 0, 1, 2 (mono, stereo, multi)
   ochn is 0, 1, 2 (mono, stereo, multi)
   ibpc and obpc are 0, 1, 3 for 8, 16 or 32 bits (indexes == 2 are NULLs)
*/
extern copy_func copy_tab[3][3][4][4];
extern resample_func resample_tab[3][3][4][4];

void select_copy_samples_func(int ichn, int ochn, int ibpc, int obpc)
{
    if(ichn >= 3) ichn = 3;
    if(ochn >= 3) ochn = 3;
    resample_samples = resample_tab[ichn-1][ochn-1][ibpc-1][obpc-1];
    copy_samples = copy_tab[ichn-1][ochn-1][ibpc-1][obpc-1];
}

copy_func copy_samples_sep = NULL;
resample_func resample_samples_sep = NULL;
/* indexes are the same as above, except for ichn, which is
   0 or 1 (stereo or multi, mono makes no sense) */
extern copy_func copy_sep_tab[2][3][4][4];
extern resample_func resample_sep_tab[2][3][4][4];

void select_copy_samples_sep_func(int ichn, int ochn, int ibpc, int obpc)
{
    if(ichn == 1)
        { resample_samples_sep=NULL; copy_samples_sep=NULL; return; }
    if(ichn >= 3) ichn = 3;
    if(ochn >= 3) ochn = 3;
    resample_samples_sep = resample_sep_tab[ichn-2][ochn-1][ibpc-1][obpc-1];
    copy_samples_sep = copy_sep_tab[ichn-2][ochn-1][ibpc-1][obpc-1];
}

/* this file is generated by m4 from output_funcs.h.m4:
   I'm not (yet) so crazy to write by hand all that bullshit ;-) */
#include "output_funcs.h"

/* Hey, these tables are eating 4*2*2*4*2 = 128 bytes! ;-))) */
convert_func convert_func_in_tab[2][2][4] = { /* [unsig][swap][bpc-1] */
    {
        { NULL, NULL, NULL, NULL }, /* no conversions */
        { convert_swap_8, convert_swap_16, NULL, convert_swap_32 }
    },
    {
        { convert_unsig_8, convert_unsig_16, NULL, convert_unsig_32 },
        { convert_swap_unsig_8, convert_swap_unsig_16, NULL,
          convert_swap_unsig_32 }
    }
};

convert_func convert_func_out_tab[2][2][4] = { /* [unsig][swap][bpc-1] */
    {
        { NULL, NULL, NULL, NULL }, /* no conversions */
        { convert_swap_8, convert_swap_16, NULL, convert_swap_32 }
    },
    {
        { convert_unsig_8, convert_unsig_16, NULL, convert_unsig_32 },
        { convert_unsig_swap_8, convert_unsig_swap_16, NULL,
          convert_unsig_swap_32 }
    }
};

int select_pre_post_convert_func(void)
{
    assert(in_bpc > 0 && in_bpc <= 4);
    assert(out_bpc > 0 && out_bpc <= 4);
    assert(oss_swapin == (oss_swapin != 0)); /* be sure it's a truth value */
    assert(oss_swapout == (oss_swapout != 0)); /* idem */
    assert(oss_unsig_in == (oss_unsig_in != 0)); /* idem */
    assert(oss_unsig_out == (oss_unsig_out != 0)); /* idem */

    convert_in = convert_func_in_tab[oss_unsig_in][oss_swapin][in_bpc-1];
    convert_out = convert_func_out_tab[oss_unsig_out][oss_swapout][out_bpc-1];
    return 1;
}

/* output_thread: feeds sound_fd with a (almost) constant data flow,
   while the buffer is filled by the main thread
   ("main" from out point of view) */

pthread_t out_thr = (pthread_t) -1;

void* output_thread(void*);

int start_output_thread(void)
{
    out_thr_msg = OMSG_RUN;
    if((int)out_thr < 0)
        return pthread_create(&out_thr, NULL, output_thread, NULL);
    else
        return 0;
}

int term_output_thread(int flush)
{
    int ret;
    pthread_mutex_lock(&buf_mtx);
    out_thr_msg = flush ? OMSG_TERM : OMSG_TERM_NOW;
    DBGMSG("term_output_thread: signalling buf_cond...\n");
    pthread_cond_broadcast(&buf_cond);
    pthread_mutex_unlock(&buf_mtx);
    ret = pthread_join(out_thr, NULL);
    if(ret == 0)
        out_thr = (pthread_t) -1;
    return ret;
}

static inline int must_term(void) {
    return out_thr_msg == OMSG_TERM || out_thr_msg == OMSG_TERM_NOW;
}

static inline int must_flush(void) {
    return out_thr_msg == OMSG_FLUSH || out_thr_msg == OMSG_TERM;
}

static inline int dont_wait(void) {
    return out_thr_msg == OMSG_FLUSH || must_term();
}

void write_to_device(int len)
{
    int n, i, ret;
    char* p;
/*     DBGMSG("write_to_device(%d): start\n", len); */
    for(; len > 0; len -= n) {
        n = buffer_read(&tmpbuf, &p);
        if(n > len) n = len;
        if(n > 0) {
            for(i = n; i > 0; i -= ret, p += ret) {
/*              DBGMSG("write_to_device: writing %d bytes\n", i); */
                ret = write(sound_fd, p, i);
                if(ret < 0) {
                    if(errno == EINTR)
                        ret = 0; /* simply need to repeat the call */
                    else {
                        fprintf(stderr, "%s: write: %s\n", device, strerror(errno));
                        n -= i;
                        break;
                    }
                }
            }
            pthread_mutex_lock(&buf_mtx);
            buffer_get(&tmpbuf, n);
            pthread_mutex_unlock(&buf_mtx);
        }
    }
    pthread_mutex_lock(&buf_mtx);
/*     DBGMSG("write_to_device: signalling buf_cond...\n"); */
    pthread_cond_broadcast(&buf_cond);
    pthread_mutex_unlock(&buf_mtx);
}

void* output_thread(void* arg)
{
    DBG(void* dbgmt = dbg_meter_new("output buffer");)
    DBG(void* dbgmtdev = dbg_meter_new("device buffer");)
    audio_buf_info binfo;
    int start = 1; /* just started, don't detect buffer underruns */
    int absize = 0; /* audio buffer size */
    /* minimum space to wait for on the device after filling it up:
       it's 15% of total kernel buffer (empirically choosen) */
    int timeout; /* time to wait for oss_minspace */
    int fragsize;
    int minbuf; /* minimum data to wait for in our buffer */
    int bufund = 0; /* buffer underruns counter */
    struct timeval tv;
    fd_set set;
    (void) arg;
    
    /* Taken from xmms 1.2.8
     * 
     * Stupid hack to find out if the driver support selects, some
     * drivers won't work properly without a select and some won't
     * work with a select :/
     */
    tv.tv_sec = 0;
    tv.tv_usec = 50000;
    FD_ZERO(&set);
    FD_SET(sound_fd, &set);
    if (select(sound_fd + 1, NULL, &set, NULL, &tv) > 0)
        select_works = 1;
    else
        select_works = 0;

    DBGMSG("output_thread: started\n");
    if(is_oss) {
        if(ioctl(sound_fd, SNDCTL_DSP_GETOSPACE, &binfo) == 0) {
            absize = binfo.fragstotal * binfo.fragsize; 
            fragsize = binfo.fragsize;
            minbuf = (int)oss_minspace < fragsize ?
                          (int)oss_minspace : fragsize;
            timeout = (int)(oss_minspace * 1000000.0 / (out_bps * out_fmt.rate));
        } else {
            /* arbitrary (but reasonable) value of 10 msec */
            minbuf = oss_minspace;
            timeout = 10000;
            fragsize = (tmpbuf.end - tmpbuf.start) / 4;
        }
    } else {
        timeout = 10000; /* meaningless */
        minbuf = 0; /* meaningless */
        fragsize = (tmpbuf.end - tmpbuf.start) / 4;
    }
    if(minbuf > buffer_size(&tmpbuf)) {
        /* safety check -- THIS SHOULDN'T HAPPEN  */
        DBGMSG("BUG: buffer is smaller tha fragment size!!!\n");
        minbuf = buffer_size(&tmpbuf);
    }
    DBGMSG("output_thread: fragsize = %d, absize = %d, minbuf = %d\n",
           fragsize, absize, minbuf);
    while(out_thr_msg != OMSG_TERM_NOW) {
        int data;

/*      DBGMSG("output_thread: top of main loop, msg = %d\n", out_thr_msg); */
        if(out_thr_msg == OMSG_PAUSE) {
            struct timespec ts;
            ts.tv_sec = 0;
            ts.tv_nsec = 10000; /* 10msec */
            nanosleep(&ts, &ts);
            start = 1;
            continue;
        }

        if(must_flush() && buffer_empty(&tmpbuf)) {
            if(is_oss)
                ioctl(sound_fd, SNDCTL_DSP_POST, NULL);
            if(out_thr_msg == OMSG_TERM)
                break;
            start = 1;
        }

        /* wait for some data in our buffer */
        pthread_mutex_lock(&buf_mtx);
        DBG(dbg_meter_show(dbgmt, buffer_data(&tmpbuf) * 100 / buffer_size(&tmpbuf));)
        if(start) { /* wait for at least one fragment at the beginning */
            while(!dont_wait() && buffer_data(&tmpbuf) < fragsize) {
/*                 DBGMSG("output_thread: buffer: %db, waiting for %db\n", */
/*                        buffer_data(&tmpbuf), fragsize); */
                pthread_cond_wait(&buf_cond, &buf_mtx);
            }
        } else {
            while(!dont_wait() && buffer_data(&tmpbuf) < minbuf) {
/*                 DBGMSG("output_thread: buffer empty, waiting for data\n"); */
                pthread_cond_wait(&buf_cond, &buf_mtx);
            }
        }
        pthread_mutex_unlock(&buf_mtx);

        data = buffer_data(&tmpbuf);

        if(is_oss && ioctl(sound_fd, SNDCTL_DSP_GETOSPACE, &binfo)==0) {
            /* if there's enough space in the device buffer, write all, else
               fill buffer and wait for minspace */
/*          DBGMSG("output_thread: device buf has %d/%d space\n", */
/*                 binfo.bytes, absize); */
            DBG(dbg_meter_show(dbgmtdev, 100 - binfo.bytes * 100 / absize);)
            if(binfo.bytes >= data) {
                int delay;
                if(is_oss) {
                    /* detect a buffer underrun */
                    if(verbose && !start &&
                       ioctl(sound_fd, SNDCTL_DSP_GETODELAY, &delay) == 0) {
                        if(delay <= 0) {
                            fprintf(stderr, "%s: buffer underrun\n", device);
                            bufund++;
                        }
                    }
                }
                write_to_device(data);
                start = 0;
            } else {
                struct timespec ts;
                write_to_device(binfo.bytes / out_bps * out_bps);
                start = 0;
                binfo.bytes = 0; /* we've just written them */
                DBG(dbg_meter_show(dbgmtdev, 100);)
                /* wait 'timeout' usec to free 'oss_minspace' bytes */
/*              DBGMSG("output_thread: waiting %d usec (till %.9f)\n", */
/*                     timeout, (double)ts.tv_sec + ts.tv_nsec * 1e-9); */
                if(!select_works) {
                    ts.tv_sec = 0;
                    ts.tv_nsec = timeout * 1000;
                    while(!dont_wait() && binfo.bytes < (int)oss_minspace) {
                        nanosleep(&ts, NULL);
                        if(ioctl(sound_fd, SNDCTL_DSP_GETOSPACE, &binfo) != 0)
                            break;
                    }
                } else {
                    FD_ZERO(&set);
                    FD_SET(sound_fd, &set);
                    tv.tv_sec = timeout / 1000000L;
                    tv.tv_usec = timeout & 1000000L;
                    select(sound_fd + 1, NULL, &set, NULL, &tv);
                }
            }
        } else {
            /* not an oss device, or errors with GETOSPACE */
            write_to_device(data);
            if(!is_oss)
                start = 1; /* so we always wait for a prebuffer */
        }
    }
    DBG(dbg_meter_delete(dbgmt);)
    DBG(dbg_meter_delete(dbgmtdev);)
    DBGMSG("output_thread: ended\n");
    return (void*) 0;
}

/* auxiliary function for do_write */
int do_copy(void* data, int nsamp, char* buffer, int nbuf, long* ni)
{
    long nc; /* number of samples copied */
    char *p, *end;

    if(convert_in != NULL) {
        /* yeah, I know that it's not wise to modify input data ... but
           convert_in is used only for testing and nothing else (by now) */
        end = (char*) data + nsamp * in_bps;
        for(p = (char*) data; p < end; p += in_bpc)
            write_sample_in(p, convert_in(read_sample(p)));
    }
    switch(conv_type) {
    case conv_none:
        nc = *ni = (nsamp < nbuf ? nsamp : nbuf);
        memcpy(buffer, data, nc * out_bps);
        break;
    case conv_copy:
        nc = (nsamp < nbuf ? nsamp : nbuf);
        nc = *ni = copy_samples(data, buffer, nc);
        break;
    case conv_resample:
        nc = resample_samples(data, buffer, nsamp, nbuf, ni, &resamp_state);
        break;
    default:
        DBGMSG("BUGBUG: do_write: unknown value for conv_type\n");
        nc = *ni = 0;
        break;
    }
    if(convert_out != NULL) {
        end = buffer + nc * out_bps;
        for(p = buffer; p < end; p += out_bpc)
            write_sample(p, convert_out(read_sample_out(p)));
    }
    return nc;
}

/* Writes nsamp bytes using buffer as temporary space (preallocated).
   Buffer is already filled with ibuf samples, when it became full,
   the buffer as a whole is written to sound_fd, bufi is modified
   accordingly.
   Returns number of input samples processed.
 */
int do_write(void* data, int nsamp, buffer *buf)
{
    long nc; /* number of samples copied */
    long ni; /* number of input samples processed (if != nc) */
    int i;
    char *p;

    if(nsamp <= 0) return 0;

    out_thr_msg = OMSG_RUN;

    pthread_mutex_lock(&buf_mtx);
    if(buffer_full(buf)) {
        /* wait for buffer to have at least 'oss_minspace' space */
        while(buffer_space(buf) < (int)oss_minspace) {
/*             DBGMSG("do_write: signalling buf_cond...\n"); */
            pthread_cond_broadcast(&buf_cond);
/*             DBGMSG("do_write: waiting for space in buffer...\n"); */
            pthread_cond_wait(&buf_cond, &buf_mtx);
        }
/*         DBGMSG("do_write: now buf has %d/%d space\n", buffer_space(buf), */
/*                buf->end - buf->start); */
    }
    pthread_mutex_unlock(&buf_mtx);

    i = buffer_write(buf, &p);
    nc = do_copy(data, nsamp, p, i / out_bps, &ni);

    DBG(if(nc == 0)
        DBGMSG("do_write: wrote zero bytes, headi=%d, taili=%d\n",
               buffer_space(&tmpbuf), tmpbuf.head-tmpbuf.start,
               tmpbuf.tail-tmpbuf.start);
    )
/*     DBGMSG("do_write: just wrote %d bytes\n", nc * out_bps); */
    pthread_mutex_lock(&buf_mtx);
    buffer_put(buf, nc * out_bps); /* advance pointer */
    pthread_mutex_unlock(&buf_mtx);

    return ni;
}

/* Write data to the output device, in the format specified
   in the call to output_open ('fmt', not 'actual_fmt'), and start
   playing it.
   'data' is the start address of sound data, and nsamp is the number
   of samples (a left-right pair counts as ONE sample).
   If a sample needs more than one byte, endianess is intended to be
   the preferred of the current architecture (i.e. the one you get with
   '*data = sample;' when sizeof(sample) > 1), check ifdef WORD_BIGENDIAN.
   If bits is not a power of 2 (>=8), it's padded to the greatest power
   of 2 when loading samples from 'data' (i.e. bits=24 ==> a sample is 32
   bits wide, 8 of which are unused).
   Channels are first left, then right with stereo output.
   As usual, return is 0 on error and != 0 on success.
   N.B. this call could block if the device buffer is not big enough.
*/
int output_write(void* data, int nsamp)
{
    if(conv_type != conv_none && tmpbuf.start == NULL) {
        DBGMSG("BUG: tmpbuf == NULL but it shouldn't\n");
        /* this should do nothing, tmpbuf is allocated in output_open */
        allocate_temp_buffer();
    }

/*     DBGMSG("output_write: writing %d samples (%d bytes)\n", */
/* 	   nsamp, nsamp*out_bps); */
    do {
        int ret = do_write(data, nsamp, &tmpbuf);
        if(ret < 0) return 0;
        nsamp -= ret;
        data += ret * in_bps;
    } while(nsamp > 0);
    pthread_mutex_lock(&buf_mtx);
/*     DBGMSG("output_write: signalling buf_cond...\n"); */
    pthread_cond_broadcast(&buf_cond); /* wake up output_thread */
    pthread_mutex_unlock(&buf_mtx);
    return 1;
}

/* this part is done maily to handle input from libmad mpeg audio decoder,
   which uses a separate buffer for each channel :(( 
   (of course preprocessing to put that data in a suitable form is a big
   time loss) */

/* write from n separated input buffers (one for each channel) */
int output_write_chans(void* data[], int nsamp)
{
    int ret = 1;
    int i;
    copy_func old_cs = copy_samples;
    resample_func old_rs = resample_samples;
    int old_ct = conv_type;
    int old_ibps = in_bps;

    if(in_fmt.chan == 1) /* trivial case */
    return output_write(data[0], nsamp);

    if(tmpbuf.start == NULL) {
        if(conv_type == conv_none)
            conv_type = conv_copy; /* we always need a tmp buffer :(( */
        allocate_temp_buffer();
    }

    if(!copy_samples_sep)
        select_copy_samples_sep_func(in_fmt.chan, out_fmt.chan,
                                     in_bpc, out_bpc);
    copy_samples = copy_samples_sep;
    resample_samples = resample_samples_sep;
    
    input_data = (char**) data;
    /* trick to force copy/resample to increment correctly 'in' */
    in_bps = in_bpc;
    i = 0;

    do {
        int ret = do_write((char*)i, nsamp, &tmpbuf);
        if(ret < 0) { ret = 0; break; }
        nsamp -= ret;
        i += ret * in_bps;
    } while(nsamp > 0);

    pthread_mutex_lock(&buf_mtx);
/*     DBGMSG("output_write_chans: signalling buf_cond...\n"); */
    pthread_cond_broadcast(&buf_cond); /* wake up output_thread */
    pthread_mutex_unlock(&buf_mtx);

    conv_type = old_ct;
    copy_samples = old_cs;
    resample_samples = old_rs;
    in_bps = old_ibps;
    return ret;
}

/* Returns informations about the buffer: in buftotal is returned the total
   size of the buffer(s) in samples, and in bufavail the free space in samples.
   If one of these two is NULL it is ignored.
   If no device is opened, this call fails (return 0) else it returns 1.
 */
int output_bufinfo(int *buftotal, int *bufavail)
{
  if(sound_fd >= 0) {
    if(buftotal)
      *buftotal = (tmpbuf.end - tmpbuf.start + oss_nfrags * oss_fragsize) / out_bps;
    if(bufavail) {
      audio_buf_info binfo;
      if(ioctl(sound_fd, SNDCTL_DSP_GETOSPACE, &binfo) == 0)
        *bufavail = (buffer_space(&tmpbuf) + binfo.bytes) / out_bps;
      else
        *bufavail = buffer_space(&tmpbuf) / out_bps;
    }
    return 1;
  } else
    return 0;
}
