/*
 * wmdimp
 * Copyright (c) 2004 Daniele Tamino <aslo@wmdimp.cjb.net>
 *
 * This software covered by the GPL.  See COPYING file for details.
 *
 * output.h
 *
 * Output module (header)
 *
 */
#ifndef outputH
#define outputH

#include "parsecfg.h"
#include "mytypes.h"

typedef struct output_fmt {
    long rate; /* sample rate in Hz (44100 for cd quality) */
    int bits; /* bits per sample (16 for cd quality)  */
    int chan; /* n. of channels (1 = mono, 2 = stereo, >2 ...  */
} output_fmt;

extern type_desc_spec output_specs; /* config option description */

/* loads global configuration options */
int output_init(opt_hash* ht);

/* global cleanup */
void output_finish(void);

/* Opens the preferred audio device, as specified in the config file.
   If format is not supported by the device, it will be emulated
   in software (downsampling, quantization, chan. merging).
   By now, downsampling is done straight (no interpolation/antialias),
   quantization is done with dithering and noise shaping,
   channel merging is done after downsampling and before quantization.
   The actual format is stored in actual_fmt (if != NULL).
   Return value is 0 on error and !=0 on success.
   N.B. By now, fmt->bits cannot be greater than 32, and channel
   merging works only from 2 (stereo) to 1 (mono). If these condition
   are not met, the call fails (returns 0).
 */
int output_open(output_fmt* fmt, output_fmt* actual_fmt);

/* Write data to the output device, in the format specified
   in the call to output_open ('fmt', not 'actual_fmt'), and start
   playing it.
   'data' is the start address of sound data, and nsamp is the number
   of samples (a left-right pair counts as ONE sample).
   If a sample needs more than one bite, endianess is intended to be
   the preferred of the current architecture (i.e. the one you get with
   '*data = sample;' when sizeof(sample) > 1), check ifdef WORD_BIGENDIAN.
   If bits is not a power of 2 (>=8), it's padded to the greatest power
   of 2 when loading samples from 'data' (i.e. bits=24 ==> a sample is 32
   bits wide, 8 of which are unused).
   Channels are first left, then right with stereo output.
   As usual, return is 0 on error and != 0 on success.
   N.B. this call could block if the device buffer is not big enough.
*/
int output_write(void* data, int nsamp);

/* write from n separated input buffers (one for each channel) */
int output_write_chans(void* data[], int nsamp);

/* stops playing immediately; do not discard audio data no yet processed.
   when called a second time, resume from where it was stopped.
   If this is not supported, this call is equivalent to output_stop(0, 1)
   or, if that isn't supported too, to output_stop(1, 0).
   This call will never block.
*/
int output_pause(void);

/* stops playing; if now == 1 stop immediately and discard unprocessed
   data, else wait for all data to be processed, then stop.
   If now == 0, data is finished processing asynchronously if async == 1,
   else the program is blocked until all data has been processed.
   >>> Maybe async isn't very useful, so I may remove it <<<
*/
int output_stop(int now, int async);

/* Close audio device; if was playing then stop as with output_stop(1, 0)
   before closing.
 */
int output_close(void);

/* Returns informations about the buffer: in buftotal is returned the total
   size of the buffer(s) in samples, and in bufavail the free space in samples.
   If one of these two is NULL it is ignored.
   If no device is opened, this call fails (return 0) else it returns 1.
 */
int output_bufinfo(int *buftotal, int *bufavail);

#endif
