/* wmdimp = Window Maker DIgital Music Player
 * Coypright (c) 2003 Daniele Tamino
 * This software covered by the GPL.  See COPYING file for details.
 *
 * parsecfg.h
 *
 * configuration file parser (header file)
 *
 */
#ifndef parsecfgH
#define parsecfgH

#include "vect.h"

typedef float cfg_real;
typedef long  cfg_integer;
typedef int   cfg_boolean;
typedef char* cfg_string;
typedef void* cfg_struct;

/*
 * Syntax for type definition:
 *  command ::= name '=' value_list ';'
 *
 *  value ::= atomic_value | atomic_value ":" default | struct
 *  atomic_value ::= "string" | "integer" | "real" | "boolean"
 *  default ::= <a string> | <an integer> | <a real number> | <a bool value>
 *  struct ::= "{" value_list "}"
 *  value_list ::= value | array | optional | value "," value_list
 *  value_list_noopt ::= value | array | value "," value_list
 *  optional ::= "[" value_list_noopt "]"
 *  array ::= "+" value
 *
 * 'default' is the default value for the optional atomic type described.
 * (it's intentionally admitted even for non optional values, as in case of
 * error or variable not present in the config file it may be useful).
 *
 * Syntax for config file:
 *  command ::= name '=' value_list ';'
 *  name ::= <a string>
 *  value ::= atomic_value | struct
 *  atomic_value ::= <a string> | <an integer> | <a real number> | <a bool value>
 *  struct ::= "{" value_list "}"
 *  value_list ::= value | value "," value_list
 *
 * Each line describes a variable, and is in the form of
 *    name = value;
 * 'name' is the name of the variable and value is, in the type definition,
 * the name of an atomic type or the description of a structured type, and
 * in the config file, the same thing except that instead of specifying types
 * you specify the values (with matching type, of course).
 *
 * Possible atomic types are:
 * - 'string': a string, surrounded by `"' if it contains spaces;
 * - 'integer': an integer value (long);
 * - 'real': a real number (float);
 * - 'boolean': a truth value (yes/no, true/false, 1/0)
 * A structure is described in this way:
 * { atomic_type [, type, ...] }
 * A type array can be specified as '+type' and means one or more
 * occurrences of that type, and can occur only at the end of a list
 * (i.e. before the closing brace).
 * Optional values can be specified putting them in brakes "[ ]".
 * In a struct, optional types cannot be followed by non optional types.
 * So, a (irrealistic) struct could be
 *    myvar = { integer:0, string, +{ float:1.0, [ boolean:false ] } }
 * (spaces are not sensitive)
 * An example value (a line that you'll find in the config file) could be:

    myvar = { 10, "hello", { 0.25, "yes" }, { 0.75 } };

 * Values are loaded into a hash which maps from variable names to
 * trees storing that variable's value (tree of values).
 * The tree's root is the topmost node; each node can be a leaf (no
 * childs, a value stored) or composite (a struct, array or optional
 * list), with a list of childs (the inner values).
 * Each node list starts with a header with no information (can be ignored
 * when browsing), then all the other childs.
 * A 'struct' has a child for each inner member, and same for the optional
 * list (the one enclosed in brackets); An 'array' can have a variable
 * number of elements, but the last is not an element, it's a 'template'
 * used internally to load each new value; while browsing it can be useful
 * because it contains the default value (to use when the list is empty).
 * Example (the same above)

   myvar: 
       |- <header> (root)
       |- integer: value=10 default=0
       |- string: value="hello" default=""
       \- array
          |- <header>
          |- struct
          |  |- <header>
          |  |- real value=0.25 default=1.0
          |  \- boolean: value=1 default=0
          |- struct
          |  |- <header>
          |  |- real value=0.25 default=1.0
          |  \- boolean: value=0 default=0 (default taken)
          \- struct (template)
             |- <header>
             |- real value=1.0 default=1.0
             \- boolean: value=0 default=0
 */

/**************************************************************\
 * Option Tree                                                *
\**************************************************************/

struct opt_tree_node;

/* common fields of all tree nodes/leaves */
typedef struct opt_tree_node {
    /* types */
    enum type_values {
        nt_null, /* no type / placeholder */
        nt_error, /* error node */
        nt_real,
        nt_integer,
        nt_boolean,
        nt_string,
        nt_composite, /* separator between atomic and structured types (don't use) */
        nt_struct,  /* { ? , ...} */
        nt_array,   /* + ? */
        nt_optional /* [ ? , ... ] */
    } type;
    /* value depending on type  */
    union {
        cfg_real r;
        cfg_integer i;
        cfg_boolean b;
        cfg_string s;
        struct opt_tree_node* child; /* first child / child list */
    } val;
    /* default value, if it's optional and atomic */
    union {
        cfg_real r;
        cfg_integer i;
        cfg_boolean b;
        cfg_string s;
    } def;
    /* links */
    struct opt_tree_node* next;
    struct opt_tree_node* parent;
} opt_tree_node;

typedef struct opt_tree {
    opt_tree_node *root;
    int visited; /* 1 = we already stored something from config here */
} opt_tree;

int opt_tree_init(opt_tree* t);
opt_tree_node* opt_tree_insert(opt_tree_node* p, int type);
void opt_tree_free_node(opt_tree_node* nd);
void opt_tree_free(opt_tree* t);
int opt_tree_dup(opt_tree_node* src, opt_tree_node* dst, opt_tree_node* dpar);
#if 0
/* enumeration */
opt_tree_node* opt_tree_cur(opt_tree_node* n);
opt_tree_node* opt_tree_start(opt_tree_node* n);
opt_tree_node* opt_tree_end(opt_tree_node* n);
opt_tree_node* opt_tree_next(opt_tree_node* n);
opt_tree_node* opt_tree_prev(opt_tree_node* n);
#endif

/**************************************************************\
 * Option Hash Table                                          *
\**************************************************************/

typedef struct opt_hash_ent {
    char* varname;
    opt_tree *tree;
    struct opt_hash_ent* next;
} opt_hash_ent;

enum { OPT_HASH_SIZE=31 };

typedef opt_hash_ent* opt_hash[OPT_HASH_SIZE];

int opt_hash_init(opt_hash* h);
int opt_hash_insert(opt_hash* ht, const char* varname, opt_tree* tree);
opt_tree* opt_hash_find(opt_hash* ht, const char* varname);
void opt_hash_free(opt_hash* ht);

/* enumerating hash entries (in a not-better-specified order) */

typedef struct opt_hash_iterator {
    opt_hash* ht;
    int bucket;
    opt_hash_ent* cur;
} opt_hash_iterator;

opt_hash_ent* opt_hash_start(opt_hash* ht, opt_hash_iterator* it);
opt_hash_ent* opt_hash_next(opt_hash_iterator* it);
opt_hash_ent* opt_hash_cur(opt_hash_iterator *it);

/*
 * Loading the config file
 */

enum load_config_error_values {
    CFGERR_NONE     = 0, /* no error */
    CFGERR_NOTFOUND = 1, /* no config file found */
    CFGERR_SYSTEM   = 2, /* a system error (see errno?) */
    CFGERR_SPECS    = 3, /* error parsing the given specs */
    CFGERR_FILE     = 4  /* error parsing the config file */
};

enum load_config_flags_values {
    CFL_IGNUNKNOWN = 1, /* ignore unknown variables (don't print a warning) */
    CFL_IGNOVERWR  = 2, /* ignore when a variable is reassigned (overwritten) */
    CFL_SEARCHWD   = 4, /* search file in the current working directory */
};

/* Loads options from a file searching for the specified options;
 * 'opts' is a null-terminated array of 'struct config_opt'.
 * 'name' specifies what file to search; from first to last:
 * - ~/.<name>
 * - /usr/local/etc/<name>
 * - /etc/<name>
 * 'flags' is a or combination of CFL_* (see above)
 * Returns 0 if all goes well, else a number > 0.
 * Return errors are:
 * - 1: no rc file found
 * - 2: system error (see errno)
 * - 3: parse error. In this case, it prints to stderr the errors it finds.
 * - 4: syntax error. In this case, it prints to stderr the errors it finds.
 * 
 */

int load_config_file(const char* name, const char* specs, opt_hash* ht,
                     int flags);

typedef struct type_desc_spec {
    char* name; /* name (to use in error strings) */
    char* spec; /* the type description */
} type_desc_spec;

/* Same as before, but here there are more type description strings (specs
   is a null-terminated array of type_desc_specs), and each one is indipendent
   from each other, but all values are loaded from the same file in the same
   hash table.
*/
int load_config_file_multi(const char* name, const type_desc_spec** specs,
             opt_hash* ht, int flags);

/* Returns a statically allocated constant string describing err
   (as returned from load_config_file) */
const char* get_cfg_err(int err);

#endif
