/*
 * wmdimp
 * Copyright (c) 2004 Daniele Tamino <aslo@wmdimp.cjb.net>
 *
 * This software covered by the GPL.  See COPYING file for details.
 *
 * player.c
 *
 * Interface for a generic player
 *
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <stdarg.h>
#include <sys/time.h> /* need gettimeofday */
#include <dirent.h>
#include <errno.h>
#include <ctype.h>
#include <pthread.h>
#include <sys/stat.h>

#include "wmgeneral.h"
#include "buttons.h"
#include "song_hash.h"
#include "parsecfg.h"
#include "utils.h"
#include "timer.h"
#include "main.h"
#include "player.h"
#include "debug.h" /* expands to nil if #defined NDEBUG */

player* curr_player = NULL;

typedef enum { sp_up, sp_down } direction;

void set_playlist(direction d);

typedef struct playlist {
    enum type_val { pl_dir, pl_file, pl_list } type;
    char* name;     /* name of the list */
    str_list* file; /* name of directory / playlist / files */
} playlist;

void remove_list(int list);
int push_list(void); /* returns the index of the new list */
playlist* next_list();
playlist* prev_list();
playlist* current_list();
void show_list_name();
void load_next_list();
void load_prev_list();
int try_play_next(void); /* try to play the next song with curr_player */
int play_next(void);
int play_prev(void);
int setup_curr(int play); /* setup curr_song, if(play) play it */
int play(void); /* play curr_song */

const type_desc_spec player_specs = { /* player config option */
    "player",
    "fileext = +string:mp3;\n" /* mp3 valid extension(s) (default: .mp3) */
    "listext = +string:m3u;\n" /* playlist valid extension(s) (default: .m3u) */
    "alwaysscroll = boolean:no;\n" /* always scroll title (never stops) */
    "sort = string:sorted;\n" /* dir files order: unsorted, sorted, reversed */
    "scrolldelay = real:0.5;\n"    /* scrolling delay in seconds */
    "random = boolean:false;\n" /* initial state of random toggle */
    "repeat = boolean:false;\n" /* initial state of repeat toggle */
    /* list of { type, name, spec } where:
     * type is one of: dir, file, list (directory, list file or inline list)
     * name is the name of the list
     * spec is the path of the directory of playlist file (type=dir or
     * type=file), or a list of mp3 files in the case of an inline list.
     */
    "lists = +{ string:'dir', string:'(cwd)', +string:'.' };\n"
};

struct playlist** lists = NULL;
int alloc_lists = 0;
int num_lists = 0;
int curr_list = 0; /* currently loaded (or to be loaded) list */
pthread_mutex_t list_mtx;

inline const char* curr_list_name(void)
{
    return lists[curr_list]->name;
}

int load_directory(const char* dirname);
int load_playlist(const char* plname);
int load_list(str_list* list);

str_list *file_ext = NULL; /* was mp3ext */
str_list *list_ext = NULL; /* was playlistext */

int num_songs = 0;
int curr_song = 0; /* position of currently playing (or to be played) song */
int curr_song_num; /* playlist number of current song
                      (!=curr_song if random_flag is true) */
/* since my_callback can be called asynchronously, we must lock data */
pthread_mutex_t song_num_mtx;

char* title = NULL;
song_pos song_length; /* valid after setup_song */
pthread_mutex_t song_mtx;

int scroll_pos = 0;
int do_scroll = 0;
pthread_mutex_t scroll_mtx;

char* song_fname = NULL;
int always_scroll = 0;
int show_time_remain = 0;

pthread_mutex_t player_mtx;

struct timeval scroll_delay = { 0, 500000 },
               show_num_delay = { 3, 0 };

int repeat_flag = 0;
int random_flag = 0;
int *rand_song_num = NULL; /* get song from position */

int stricmp(const char* s1, const char* s2)
{
    while(*s1 != '\0' && tolower(*s1) == tolower(*s2)) {
        s1++;
        s2++;
    }
    return (tolower(*s1) - tolower(*s2));
}

/* interaction with XWindow  */

void redraw_listbar(void)
{ /* locks: display, list_mtx, song_num_mtx, song_mtx, player_mtx */
    if(is_stopped()) {
        show_list_name();
        button_redraw(B_DIR);
    } else {
        show_time();
    }
}

inline void redraw_titlebar(void)
{ /* locks: display, song_mtx, scroll_mtx */
    show_title();
}

volatile int show_msg_tit = -1; /* >=0 = showing a message in the titlebar */
volatile int show_msg_lst = -1; /* >=0 = showing a message in the listbar */

static void unshow_titlebar_hnd(struct timeval* tv, void* param)
{
    (void) param; (void) tv; /* stop compiler warnings */
    show_msg_tit = -1;
    redraw_titlebar();
    XLockDisplay(display);
    XFlush(display);
    XUnlockDisplay(display);
}
static void unshow_listbar_hnd(struct timeval* tv, void* param)
{
    (void) param; (void) tv; /* stop compiler warnings */
    show_msg_lst = -1;
    redraw_listbar();
    XLockDisplay(display);
    XFlush(display);
    XUnlockDisplay(display);
}

static int __show_message_vl(unsigned long us, int widget,
                             volatile int* show_msg, timer_handler hnd,
                             const char* fmt, va_list vl)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    tv.tv_usec += us;
    while(tv.tv_usec >= 1000000) {
        tv.tv_sec++;
        tv.tv_usec -= 1000000;
    }
    /* tv is "now + us" */
    if(*show_msg >= 0) {
        if(timer_changetime(*show_msg, &tv) < 0) {
            *show_msg = -1;
        }
    }
    if(*show_msg < 0) {
        *show_msg = timer_schedule(&tv, hnd, NULL);
    }
    return wgt_vprintf(widget, fmt, vl);
}

int show_message_vl(enum msg_loc where, unsigned long us,
                    const char* fmt, va_list vl)
{
    switch(where) {
    case msg_titlebar:
        return __show_message_vl(us, B_TITLE, &show_msg_tit,
                                 unshow_titlebar_hnd, fmt, vl);
    case msg_listbar:
        return __show_message_vl(us, B_LISTNAME, &show_msg_lst,
                                 unshow_listbar_hnd, fmt, vl);
    }
    return 0;
}

int show_message(enum msg_loc where, unsigned long us,
                 const char* fmt, ...)
{
    int ret;
    va_list vl;
    va_start(vl, fmt);
    ret = show_message_vl(where, us, fmt, vl);
    va_end(vl);
    return ret;
}

void unshow_message(enum msg_loc where)
{
    volatile int *show_msg = NULL;

    switch(where) {
    case msg_titlebar:
        show_msg = &show_msg_tit;
        break;
    case msg_listbar:
        show_msg = &show_msg_lst;
        break;
    }

    if(*show_msg >= 0) {
        timer_cancel(*show_msg);
        *show_msg = -1;
    }
}

int message_showing(enum msg_loc where)
{
    switch(where) {
    case msg_titlebar:
        return (show_msg_tit >= 0);
    case msg_listbar:
        return (show_msg_lst >= 0);
    }
    return 0;
}

void set_title(const char* tit)
{
    int len;
    DBGMSG("set_title: setting title to '%s'\n", tit);
    len = strlen(tit);
    if(title) free(title);
    title = malloc(len + 3);
    if(title) {
        memcpy(title, tit, len);
        title[len] = title[len + 1] = ' '; /* add 2 spaces at the end */
        title[len + 2] = '\0'; /* null-terminate the string */
        if(scroll_pos > len + 2) {
            if(do_scroll) scroll_pos = len + 2;
            else          scroll_pos = 0;
        }
    } else {
        /* Buy A Lot Of Rammmmmmmm (this should never happen) */
        fprintf(stderr, "BALORAMMMM (= %s)\n", strerror(ENOMEM));
    }
}

void draw_time(long min, long sec, int remain);
void draw_progress_bar(int pos, int len);

void do_update(player* pl) /* locks: display, song_num_mtx, song_mtx, player_mtx */
{
    song_pos pos, len;
    pl->get_pos(pl, &pos);
    pl->get_length(pl, &len);
    if(pos.frames > 0 && len.frames > 0) {
/*      DBGMSG("do_update: drawing progress bar\n"); */
        draw_progress_bar(pos.frames, len.frames);
        if(show_msg_lst < 0) {
            long min, sec;
/*          DBGMSG("do_update: drawing time\n"); */
            float seconds = show_time_remain ?
                len.seconds-pos.seconds : pos.seconds;
            min = (long)(seconds / 60);
            sec = (long)seconds - min * 60;
            draw_time(min, sec, show_time_remain);
        }
    } else {
        draw_progress_bar(0, 1);
        if(show_msg_lst < 0)
            draw_time(0, 0, 0);
    }
    XLockDisplay(display);
    XFlush(display);
    XUnlockDisplay(display);
}

/* facilities to re-call my_callback in another thread */

static int cb_async = 0;
pthread_cond_t cb_async_cond = PTHREAD_COND_INITIALIZER;
static inline void wait_cb(void) {
    pthread_mutex_lock(&callback_mtx);
    while(cb_async > 0)
        pthread_cond_wait(&cb_async_cond, &callback_mtx);
    pthread_mutex_unlock(&callback_mtx);
}

typedef struct cb_data {
    player_cb_func cb;
    player* pl;
    int what;
    void* param;
    void* arg;
} cb_data;

void* call_cb_thread(void* param)
{
    cb_data* d = (cb_data*) param;
    d->cb(d->pl, d->what, d->param, d->arg);
    if(d->arg) free(d->arg);
    free(d);
    pthread_mutex_lock(&callback_mtx);
    cb_async--;
    pthread_cond_broadcast(&cb_async_cond);
    pthread_mutex_unlock(&callback_mtx);
    return NULL;
}

void call_cb_async(player_cb_func cb, player* pl, int what,
                   void* param, void* arg)
{
    int err;
    pthread_t thr;
    cb_data* data;
    data = malloc(sizeof(cb_data));
    if(!data) { fputs("BALORAM=BuyALotOfRam\n", stderr); return; }
    data->cb = cb;
    data->pl = pl;
    data->what = what;
    data->param = param;
    data->arg = arg ? strdup(arg) : NULL;
    cb_async++;
    err = pthread_create(&thr, NULL, call_cb_thread, data);
    if(err != 0) {
        free(data);
        fprintf(stderr, "call_cb_async (create thread): %s\n", strerror(err));
    } else
        pthread_detach(thr);
}

/* player callback */

static int last_secs = -1;
pthread_mutex_t callback_mtx;

/* to respect mutex priority (avoid deadlocks) we mustn't lock anything
   in here, and if we need to we spawn another thread with call_cb_async */
int my_callback(struct player* pl, int what, void* async, void* arg)
{ /* locks: nothing (in the calling thread) */
    song_pos curr_pos;
    if(!async) {
        if(what == CB_SONGCHANGE || what == CB_TITCHANGE) {
            call_cb_async(my_callback, pl, what, (void*) 1, arg);
            return 1;
        }
    }
    switch(what) {
    case CB_STATUS:
        DBGMSG("my_callback: status change (%d)\n", pl->status);
        last_secs = -1.0;
        do_update(pl);
        break;
    case CB_UPDATE:
        /* update progress bar every second */
        curr_player->get_pos(curr_player, &curr_pos);
        if((int)curr_pos.seconds != last_secs) {
            last_secs = (int)curr_pos.seconds;
            do_update(pl);
        }
        break;
    case CB_SONGCHANGE:
        DBGMSG("my_callback: playing next song\n");
        play_next();
        DBGMSG("showing list num for 0.5 sec\n");
        show_list_number(500000); /* half a second */
        break;
    case CB_TITCHANGE: {
        struct hash_elt *song;
        char* s;
        DBGMSG("my_callback: title changed to \"%s\"\n", arg);
        s = strdup((char*)arg);
        if(!s) { fprintf(stderr, "BALORAMM\n"); break;  }
        pthread_mutex_lock(&song_mtx);
        song = get_song(curr_song_num);
        if(song) {
            if(song->title)
                free(song->title);
            song->title = s;
            song->info = 1;
        } else {
            DBGMSG("TITCHANGE: curr_song_num = %d is invalid!\n",
                   curr_song_num);
        }
        set_title(s);
        if(!do_scroll)
            show_title();
        pthread_mutex_unlock(&song_mtx);
        break;
    }
    case CB_ERROR:
        fprintf(stderr, "%s: %s\n", pl->name, (char*)arg);
        break;
    }
    return 1;
}

/* list loading stuff */

int (*song_sort)(const void* a, const void* b) = alphasort;
int song_sort_reverse(const void* a, const void* b) {
    return -alphasort(a, b);
}
int song_sort_none(const void* a, const void* b) {
    (void) a; (void) b; return 1;
}

int song_select(const struct dirent* entry)
{
    char* ext = getext(entry->d_name);
    str_list* ent;
    if(*ext == '.') ext++;
    for(ent = file_ext; ent; ent = ent->next)
        if(strcmp(ext, ent->str) == 0)
            return 1;
    return 0;
}

int load_directory(const char* directory)
{
    struct dirent **l = NULL;
    int i, dirlen;

    if(!directory) return ENOENT;
    dirlen = strlen(directory);

    num_songs = 0;
    empty_hash();
    num_songs = scandir(directory, &l, song_select, song_sort);
    if(num_songs < 0) {
        num_songs = 0;
        return errno;
    }
    /* store list into the hash */
    /* sorry Patrick, but why did you use a hash to store the songs?
     * isn't a normal array faster and simpler?
     */
    for(i = 0; i < num_songs; i++) {
        char* tmp = (char*) malloc(dirlen + strlen(l[i]->d_name) + 2);
        if(tmp) {
            sprintf(tmp, "%s/%s", directory, l[i]->d_name);
            insert_song(i, l[i]->d_name, tmp);
            free(tmp);
        } else { /* should never, NEVER happen */
            empty_hash();
            num_songs = 0;
            return ENOMEM;
        }
        free(l[i]);
    }
    if(l) free(l);
    return 0;
}

char* get_dir(const char* path)
{
    const char *s;
    char *dest;
    int slen = strlen(path), dlen;
    for(s = path + slen - 1; s >= path; s--)
        if(*s == '/')
            break;
    dlen = s - path + 1;
    if(dlen == 0)
        return NULL;
    dest = (char*) malloc((dlen + 1) * sizeof(char));
    if(!dest) return NULL;
    strncpy(dest, path, dlen);
    dest[dlen] = '\0';
    return dest;
}

int check_file_and_add(char* fn, char* dir)
{
    struct stat st;
    int err = 0;
    int fnlen = strlen(fn);
    int dirlen = dir ? strlen(dir) : 0;
    if(stat(fn, &st) != 0) {
        if(dir && fn[0] != '/') {
            int len = fnlen + dirlen + 1;
            char* tmp = (char*) malloc(len*sizeof(char));
            if(!tmp) return ENOMEM;
            strncpy(tmp, dir, dirlen);
            strncpy(tmp + dirlen, fn, fnlen);
            tmp[len-1] = '\0';
            if(stat(tmp, &st) == 0)
                insert_song(num_songs++, getname(tmp), tmp);
            else {
                fprintf(stderr, "Error: '%s': %s\n", tmp, strerror(errno));
                err = errno;
            }
            free(tmp);
        }
        else {
            fprintf(stderr, "Error: '%s': %s\n", fn, strerror(errno));
            err = errno;
        }
    } else
        insert_song(num_songs++, getname(fn), fn);
    return err;
}

int load_playlist(const char* file)
{
    static char buf[512];
    char *dir;
    int err = 0;
    FILE* fp;

    fp = fopen(file, "rt");
    if(!fp) return errno;
    dir = get_dir(file);
    num_songs = 0;
    empty_hash();
    while(1) {
        char* s = buf;
        fgets(buf, sizeof(buf), fp);
        if(feof(fp) || ferror(fp))
            break;
        while(*s && isspace(*s)) s++;
        stripspace(s);
        if(s[0] != '\0' && s[0] != '#') {
            int err = check_file_and_add(s, dir);
            if(err == ENOMEM) {
                return ENOMEM; /* should never happen */
            }
        }
    }
    if(dir) free(dir);
    if(ferror(fp))
        err = errno;
    fclose(fp);
    return err;
}

int load_list(str_list* file)
{
    num_songs = 0;
    empty_hash();
    for(; file != NULL; file = file->next)
        check_file_and_add(file->str, NULL);
    return 0;
}

/*
 * patch received from Steven Jorgensen to fix following function 
 */

/* up: 1 = search upward, 0 = search downward */
void set_playlist(direction d)
{ /* locks: list_mtx, song_num_mtx */
    int i, tempnum, temppos;
    int err = 0;
    int start_list;
    struct playlist *l;

    pthread_mutex_lock(&list_mtx);
    curr_song = 0; /* needed at the beginning */
    if(num_lists == 0) {
        pthread_mutex_unlock(&list_mtx);
        return;
    }
    pthread_mutex_lock(&song_num_mtx);
    start_list = curr_list;
    do {
        l = lists[curr_list];
        if(!l) break;
        pthread_mutex_lock(&song_mtx);
        switch(l->type) {
        case pl_dir:
            err = load_directory(l->file->str);
            break;
        case pl_file:
            err = load_playlist(l->file->str);
            break;
        case pl_list:
            err = load_list(l->file);
            break;
        default:
            DBG(
            fprintf(stderr, "BUG: invalid list type %d\n", l->type);
            err = EINVAL;
            )
            break;
        }
        pthread_mutex_unlock(&song_mtx);
        if(err != 0) {
            if(err != ENOENT) {
                fprintf(stderr, "Warning: removing list '%s': %s\n",
                        l->name, strerror(err));
                remove_list(curr_list);
                if(curr_list < start_list)
                    start_list--;
            } else {
                DBGMSG("Warning: skipping unavailable list '%s'\n", l->name);
                if(d == sp_up) curr_list--;
                else           curr_list++;
            }
        } else if(num_songs == 0) {
            DBGMSG("Warning: skipping empty list '%s'\n", l->name);
            if(d == sp_up) curr_list--;
            else           curr_list++;
            err = 1;
        }
        if(curr_list >= num_lists)
            curr_list = 0;
        else if(curr_list < 0)
            curr_list = num_lists - 1;
    } while(err != 0 && curr_list != start_list);

    /* Create Pseudo-random permutation of list */
    if(rand_song_num) free(rand_song_num);
    srand(time(NULL));
    rand_song_num = (int *)malloc(sizeof(int)*num_songs);
    if (rand_song_num) {
        for (i=0; i<num_songs; i++) rand_song_num[i] = i;
        for (i=0; i<num_songs; i++) {
            tempnum = rand_song_num[i];
            temppos = rand()%num_songs;
            rand_song_num[i] = rand_song_num[temppos];
            rand_song_num[temppos] = tempnum;
        }
    } else { /* This shouldn't happen - the list isn't that big */
        fprintf(stderr, "Error: cannot allocate randomized list\n");
    }
    curr_song_num = random_flag ? rand_song_num[0] : 0;
    pthread_mutex_unlock(&song_num_mtx);
    pthread_mutex_unlock(&list_mtx);
}

int do_setup_song(struct hash_elt* song, int play) /* aux function */
{
    int ret;
    char* tit = NULL;
    char** ptit = NULL;
    if(song->info == 0)
        ptit = &tit;
    ret = curr_player->setup(curr_player, song->filename, ptit, play);
    if(!song->info && ret && tit) {
        if(song->title)
            free(song->title);
        song->title = tit;
        song->info = 1;
    }
    return ret;
}

int setup_song(int i, int play, int testsong)
{ /* locks: everything in player.c, display */
    struct hash_elt *song;
    int ret = 0;
    char* s;

    if(random_flag) {
        pthread_mutex_lock(&list_mtx);
        i = rand_song_num[i]; /* get song num from position */
        pthread_mutex_unlock(&list_mtx);
    }
    curr_song_num = i; /* assumption: this is atomic */
    pthread_mutex_lock(&song_mtx);
    song = get_song(i);

    if (song) {
        pthread_mutex_lock(&player_mtx);
        /* first try to use current player */
        if(curr_player && curr_player->test(curr_player, song->filename))
           ret = do_setup_song(song, play);
        if(!ret && !testsong) {
            /* search a suitable player for this song */
            player_factory** pf;
            if(curr_player) {
                player_destroy(curr_player);
                curr_player = NULL;
            }
            for(pf = player_tab; *pf != NULL; pf++) {
                player* p = (*pf)->create();
                if(p && p->test(p, song->filename)) {
                    p->set_callback(p, my_callback, NULL);
                    curr_player = p;
                    break;
                }
                player_destroy(p);
            }
        }
        pthread_mutex_unlock(&player_mtx);
        if(!ret && testsong) return 0;
        pthread_mutex_lock(&player_mtx);
        if(curr_player) {
            if(!ret)
                ret = do_setup_song(song, play);
            if(!ret) {
                fprintf(stderr, "Error: failed to load %s with %s\n",
                        song->filename, curr_player->name);
            }
        } else {
            fprintf(stderr, "Error: %s is not in a supported format!\n",
                    song->filename);
        }
        pthread_mutex_unlock(&player_mtx);
        /* set the title */
        pthread_mutex_lock(&scroll_mtx);

        if(ret) s = song->title;
        else    s = getname(song->filename);

        do_scroll = 0;
        scroll_pos = 0;
        set_title(s);
            
        song_fname = song->filename;
        song_length.frames = -1;
        
        pthread_mutex_unlock(&scroll_mtx);

        /* scroll the title at least once */
        show_title();
        turn_on_scroll();
    }
    pthread_mutex_unlock(&song_mtx);
    return ret;
}

void go_next(void) /* locks: song_num_mtx, list_mtx */
{
    pthread_mutex_lock(&song_num_mtx);
    if (++curr_song >= num_songs) {
        if (repeat_flag)
            curr_song = 0;
        else {
            pthread_mutex_unlock(&song_num_mtx);
            load_next_list();
            pthread_mutex_lock(&song_num_mtx);
        }
    }
    pthread_mutex_unlock(&song_num_mtx);
}

int setup_curr(int play)
{
    int song_start, list_start;
    int ret;
    pthread_mutex_lock(&song_num_mtx);
    song_start = curr_song;
    list_start = curr_list;
    do {
        ret = setup_song(curr_song, play, 0);
        if(!ret)
            go_next();
    } while(!ret && (curr_list != list_start || curr_song != song_start));
    pthread_mutex_unlock(&song_num_mtx);
    return ret;
}

int setup_next(int play) /* locks: everything in player.c */
{
    int song_start, list_start;
    int ret;
    pthread_mutex_lock(&song_num_mtx);
    song_start = curr_song;
    list_start = curr_list;
    do {
        go_next();
        pthread_mutex_unlock(&song_num_mtx);
        ret = setup_song(curr_song, play, 0);
        pthread_mutex_lock(&song_num_mtx);
    } while(!ret && (curr_list != list_start || curr_song != song_start));
    pthread_mutex_unlock(&song_num_mtx);
    return ret;
}

/* try to setup and play the next song only with the current player, without
   destroying it to try the others if this fails */
int try_play_next(void)
{
    int ret;
    go_next();
    ret = setup_song(curr_song, 1, 1);
    return ret;
}

int play_next(void) { return setup_next(1); } /* locks: see setup_next above */

int play_prev(void) /* locks: everything in player.c */
{
    int song_start, list_start, i;
    int ret;

    pthread_mutex_lock(&list_mtx);
    pthread_mutex_lock(&song_num_mtx);
    song_start = i = curr_song;
    list_start = curr_list;
    pthread_mutex_unlock(&song_num_mtx);
    do {
        if (--i < 0) {
            if(!repeat_flag)
                load_prev_list();
            i = num_songs - 1;
        }
        ret = setup_song(i, 1, 0);
    } while(!ret && (curr_list != list_start || i != song_start));
    curr_song = i;
    pthread_mutex_unlock(&list_mtx);
    return ret;
}

/* play current song */
int play(void)
{
    return setup_curr(1);
}

int user_play() /* locks: everything, display */
{
    if (!is_playing()) {
        if(is_paused()) {
            return user_pause();
        } else {
            int ret;
            ret = play();
            if(ret) {
                /* hide every message and show time */
                unshow_message(msg_titlebar);
                unshow_message(msg_listbar);
                do_update(curr_player);
            }
            return ret;
        }
    }
    return 1; /* if it's already playing it's considered success */
}


int user_stop() /* (was 'stop') locks: everything */
{
    int ret = 1; /* already stopped is considered success */
    pthread_mutex_lock(&player_mtx);
    if(!curr_player) ret = 0;
    else if(curr_player->status != st_stop) {
        if(curr_player->stop(curr_player)) {
            ret = 1;
        } else
            ret = 0;
    }
    pthread_mutex_unlock(&player_mtx);
    /* actually, I find pleasing when the progressbar remains at the position
       where I've stopped the song, so I can restart easily from there :-)
       So i've disabled this update */
    /* do_update(); */
    return ret;
}

int user_next(void) /* locks: everything */
{
    if(is_playing() || is_paused()) {
        return play_next();
    } else {
        int ret;
        pthread_mutex_lock(&list_mtx);
        pthread_mutex_lock(&song_num_mtx);
        if(++curr_song >= num_songs) {
            if(repeat_flag)
                curr_song = 0;
            else {
                pthread_mutex_unlock(&song_num_mtx);
                load_next_list();
                pthread_mutex_lock(&song_num_mtx);
            }
        }
        ret = setup_song(curr_song, 0, 0);
        pthread_mutex_unlock(&song_num_mtx);
        pthread_mutex_unlock(&list_mtx);
        return ret;
    }
}

int user_back(void) /* locks: everything */
{
    if(is_playing() || is_paused()) {
        return play_prev();
    } else { 
        int ret;
        pthread_mutex_lock(&list_mtx);
        pthread_mutex_lock(&song_num_mtx);
        if(--curr_song < 0) {
            if(repeat_flag)
                curr_song = num_songs - 1;
            else {
                pthread_mutex_unlock(&song_num_mtx);
                load_prev_list();
                pthread_mutex_lock(&song_num_mtx);
                curr_song = num_songs - 1;
            }
        }
        ret = setup_song(curr_song, 0, 0);
        pthread_mutex_unlock(&song_num_mtx);
        pthread_mutex_unlock(&list_mtx);
        return ret;
    }
}

int user_jump(long progress, long total) /* was 'jump' */
{
    if(!is_playing()) {
        if(!setup_curr(1)) return 0;
    }
    return curr_player->jump(curr_player, progress, total);
}

int user_jump_frame(long frame)
{
  if(!is_playing()) {
      if(!setup_curr(1)) return 0;
  }
  return curr_player->jump_frames(curr_player, frame);
}

int user_jump_time(long sec)
{
  if(!is_playing()) {
      if(!setup_curr(1)) return 0;
  }
  return curr_player->jump_time(curr_player, sec);
}

/*
 * initialization functions
 */

void player_init(opt_hash* ht) 
{
    opt_tree* t;
    pthread_mutexattr_t mtx_attr;
    /* inizialize lists */
    lists = malloc(sizeof(playlist*) * 8);
    if(!lists) {
        fprintf(stderr, "You are very low on memory!!!\n");
        exit(-1);
    }
    alloc_lists = 8;

    /* all mutexes are better to be recursive */
    pthread_mutexattr_init(&mtx_attr);
    pthread_mutexattr_settype(&mtx_attr, PTHREAD_MUTEX_RECURSIVE_NP);
    pthread_mutex_init(&list_mtx, &mtx_attr);
    pthread_mutex_init(&song_num_mtx, &mtx_attr);
    pthread_mutex_init(&song_mtx, &mtx_attr);
    pthread_mutex_init(&scroll_mtx, &mtx_attr);
    pthread_mutex_init(&player_mtx, &mtx_attr);
    pthread_mutex_init(&callback_mtx, &mtx_attr);
    pthread_mutexattr_destroy(&mtx_attr);

    /* load config options */

    t = opt_hash_find(ht, "fileext");
    assert(t && t->root && t->root->next);
    set_file_ext(t->root->next);

    t = opt_hash_find(ht, "listext");
    assert(t && t->root && t->root->next);
    set_list_ext(t->root->next);

    t = opt_hash_find(ht, "alwaysscroll");
    assert(t && t->root && t->root->next);
    set_alwaysscroll(t->root->next->val.b);

    t = opt_hash_find(ht, "sort");
    assert(t && t->root && t->root->next);
    set_sorting(t->root->next->val.s);

    t = opt_hash_find(ht, "scrolldelay");
    assert(t && t->root && t->root->next);
    set_scrolldelay(t->root->next->val.r);

    t = opt_hash_find(ht, "lists");
    assert(t && t->root && t->root->next);
    set_lists(t->root->next);

    t = opt_hash_find(ht, "random");
    assert(t && t->root && t->root->next);
    random_flag = t->root->next->val.b;

    t = opt_hash_find(ht, "repeat");
    assert(t && t->root && t->root->next);
    repeat_flag = t->root->next->val.b;

    set_playlist(sp_down);
    static_init_all_players(ht);
    timer_start();
}

void free_list(playlist* l)
{
    if(l->name) free(l->name);
    if(l->file) free_str_list(l->file);
    free(l);
}

void player_finish(void)
{
    wait_cb();
    if(curr_player) {
        player_destroy(curr_player);
        curr_player = NULL;
    }
    static_finish_all_players();
    pthread_mutex_unlock(&callback_mtx);
    if(file_ext)
        free_str_list(file_ext);
    if(list_ext)
        free_str_list(list_ext);
    if(lists) {
        int i;
        for(i = 0; i < num_lists; ++i)
            free_list(lists[i]);
        free(lists);
    }
    if(rand_song_num) free(rand_song_num);
    if(title) free(title);
    timer_stop();
    empty_hash(); /* useful when tracing malloc */
}

/*
 * Configuration file options
 */

#define ASSERT_MALLOC(x) if(!(x)) { fprintf(stderr, "BUY MORE RAMM"); return; }

void set_file_ext(opt_tree_node* n)
{
    load_str_list(n, &file_ext);
}

void set_list_ext(opt_tree_node* n)
{
    load_str_list(n, &list_ext);
}

void set_alwaysscroll(int as) {
    always_scroll = as;
}

void set_scrolldelay(float sd)
{
    if(sd < 0.02) {
        fprintf(stderr, "Warning: scrolldelay %g < 0.02s may overload the system!",
                sd);
    }
    if(sd < 0.000001) {
        fprintf(stderr, "Warning: scrolldelay is %g < 1 microsec"
                "(using 1 us)\n", sd);
        scroll_delay.tv_sec = 0;
        scroll_delay.tv_usec = 1;
        return;
    }
    if(sd < 0) {
        fprintf(stderr, "Error: scrolldelay is %g, should be > 0"
                "(using absolute value)\n", sd);
        sd = -sd;
    }
    scroll_delay.tv_sec  = (unsigned long)sd;
    scroll_delay.tv_usec = (unsigned long)((sd - scroll_delay.tv_sec)
                                           * 1000000);
}

void set_shownumdelay(float sd)
{
    if(sd < 0.02) {
        fprintf(stderr, "Warning: shownumdelay %g < 0.02s may overload the system!",
                sd);
    }
    if(sd < 0.000001) {
        fprintf(stderr, "Warning: shownumdelay is %g < 1 microsec"
                "(using 1 us)\n", sd);
        show_num_delay.tv_sec = 0;
        show_num_delay.tv_usec = 1;
        return;
    }
    if(sd < 0) {
        fprintf(stderr, "Error: shownumdelay is %g, should be > 0"
                "(using absolute value)\n", sd);
        sd = -sd;
    }
    show_num_delay.tv_sec  = (unsigned long)sd;
    show_num_delay.tv_usec = (unsigned long)((sd - show_num_delay.tv_sec)
                                             * 1000000);
}

void set_lists(opt_tree_node* array) /* locks: list_mtx, song_num_mtx */
{
    opt_tree_node* n;
    playlist* lst;
    int use_default = 0;
    int i;
    if(!array) return;
    n = array->val.child;
    if(!n) return;
    pthread_mutex_lock(&list_mtx);
    /* check if list is empty (only template), and use defaults in
       this case */
    if(n->next->next == NULL)
        use_default = 1;
    /* skip first (header) and last (template) elements (if not the
       only one) */
    for(n = n->next; n && (n->next || use_default); n = n->next) {
        opt_tree_node* pl = n->val.child->next;
        lst = lists[i = push_list()];
        lst->name = strdup(pl->next->val.s);
        lst->file = NULL;
        if(strncmp(pl->val.s, "dir", 4) == 0) {
            lst->type = pl_dir;
        } else if(strncmp(pl->val.s, "file", 4) == 0) {
            lst->type = pl_file;
        } else if(strncmp(pl->val.s, "list", 4) == 0) {
            lst->type = pl_list;
        } else {
            fprintf(stderr, "Error: list '%s': unknown type '%s'.\n",
                    lst->name, pl->val.s);
            remove_list(i);
            continue;
        }
        load_str_list(pl->next->next, &lst->file);
    }
    curr_list = 0; /* assumption: this is atomic */
    set_playlist(sp_down);
    pthread_mutex_unlock(&list_mtx);
}

void set_sorting(char *s)
{
    char *temp = s;
    while (*temp) {
        *temp = tolower(*temp);
        temp++;
    }
    if ( !strcmp(s, "sorted") || !strcmp(s, "on") || !strcmp(s, "yes") ||
         !strcmp(s, "1") || !strcmp(s, "true")) {
        song_sort = alphasort;
    } else if ( !strcmp(s, "unsorted") || !strcmp(s, "off") || !strcmp(s, "no")
                || !strcmp(s, "0") || !strcmp(s, "false")) {
        song_sort = song_sort_none;
    } else if ( !strcmp(s, "reversed")) {
        song_sort = song_sort_reverse;
    } else {
        fprintf(stderr, "Unknown sorting option: '%s'\n", s);
    }
}

/* returns the index of the new list */
int push_list(void)
{
    if(num_lists >= alloc_lists) { /* expand */
        playlist** l = realloc(lists, (alloc_lists * 2) * sizeof(playlist*));
        if(!l) {
            fprintf(stderr, "You are very low on memory!!!\n");
            return /* -1 */ num_lists;
        }
        lists = l;
        alloc_lists *= 2;
    }
    lists[num_lists] = (playlist*) malloc(sizeof(playlist));
    return num_lists++;
}

void remove_list(int list)
{
    int i;
    if(list < 0 || list >= num_lists) return;
    if(lists[list])
        free_list(lists[list]);
    for(i = list; i < num_lists - 1; i++)
        lists[i] = lists[i + 1];
    num_lists--;
    if(alloc_lists >= 4*num_lists) { /* shrink */
        playlist** l = realloc(lists, (alloc_lists / 2) * sizeof(playlist*));
        if(l) {
            lists = l;
            alloc_lists /= 2;
        }
    }
}

/*
 * list manipulation
 */

playlist* next_list()
{
    if (++curr_list >= num_lists)
        curr_list = 0;
    return (lists[curr_list]);
}

playlist* prev_list()
{
    if (--curr_list < 0)
        /* protect against num_lists being <= 0 */
        curr_list = num_lists - 1 > 0 ? num_lists - 1 : 0;
    return (lists[curr_list]);
}

playlist* current_list()
{
    return (lists[curr_list]);
}

void show_list_name() /* locks: display, list_mtx */
{
    struct coord* dirpos = &btn_pos[B_DIR];
    int cl;

    pthread_mutex_lock(&list_mtx);
    pthread_mutex_lock(&song_num_mtx);
    cl = curr_list;
    pthread_mutex_unlock(&song_num_mtx);
    if (lists[cl] && lists[cl]->name != NULL) {
        int i, len, width = dirpos->w / char_width;
        char* tmp = (char*) malloc(width + 1);
        strncpy(tmp, lists[cl]->name, width);
        pthread_mutex_unlock(&list_mtx);
        len = strlen(tmp);
        for(i = len; i < width; i++)
            tmp[i] = ' '; /* fill with spaces */
/*      DBGMSG("show_list_name: showing '%s'\n", tmp); */
        draw_string(tmp, dirpos->x, dirpos->y, dirpos->w);
        free(tmp);
    } else {
        pthread_mutex_unlock(&list_mtx);
/*      DBGMSG("show_list_name: no list name (showing 'no name')\n"); */
        draw_string("no name", dirpos->x, dirpos->y, dirpos->w);
    }
    button_redraw(B_LISTNAME);
}

void list_up(/* int button_num */) /* locks: display, everything in player.c */
{
    if (!is_playing() && !is_paused()) {
        load_prev_list();
        /* after loading a list, curr_song is always == 0 */
        setup_song(0, 0, 0);
        show_list_name();
    }
}

void list_down(/* int button_num */) /* locks: display, everything in player.c */
{
    if (!is_playing() && !is_paused()) {
        load_next_list();
        setup_song(0, 0, 0);
        show_list_name();
    }
}

void load_next_list(void) /* locks: list_mtx */
{
    pthread_mutex_lock(&list_mtx);
    next_list();
    pthread_mutex_unlock(&list_mtx);
    set_playlist(sp_down);
}

void load_prev_list(void) /* locks: list_mtx */
{
    pthread_mutex_lock(&list_mtx);
    prev_list();
    pthread_mutex_unlock(&list_mtx);
    set_playlist(sp_up);
}

/* utility to draw a string on a widget (title bar, listname bar) */

#if !HAVE_VSNPRINTF
#error "need vsnprintf, get a C99 compiler (or write vsnprintf yourself)"
#endif

int wgt_vprintf(int widget, const char* fmt, va_list vl) /* locks: display */
{
    int ret, i;
    struct coord* dpos = &btn_pos[widget];
    int widget_len = dpos->w / char_width;
    char *buffer;

    if(widget_len <= 0)
        return 0;
    buffer = malloc(sizeof(char) * (widget_len + 1));

    if(!buffer) {
        fprintf(stderr, "You are on VERY low memory!!!\n");
        return 0;
    }

    ret = vsnprintf(buffer, widget_len + 1, fmt, vl);
    if(ret < widget_len) {
        for(i = ret; i < widget_len; i++)
            buffer[i] = ' ';
        buffer[widget_len] = '\0';
    }

/*  DBGMSG("wgt_printf: widget %d: showing '%s'\n", widget, buffer); */
    draw_string(buffer, dpos->x, dpos->y, dpos->w);
/*  DBGMSG("wgt_printf: redrawing\n"); */
    RedrawRegion(dpos->x, dpos->y, dpos->w, dpos->h);
    free(buffer);
    return ret;
}

int wgt_printf(int widget, const char* fmt, ...) /* locks: display */
{
    int ret;
    va_list vl;
    va_start(vl, fmt);
    ret = wgt_vprintf(widget, fmt, vl);
    va_end(vl);
    return ret;
}

/*
 * song title functions
 */

void show_title(void) /* locks: display, song_mtx, scroll_mtx */
{
    int title_len = 0;
    struct coord* tpos = &btn_pos[B_TITLE];
    int titlebar_len = tpos->w / char_width;
    char *s = "";
    int must_free = 0;

    pthread_mutex_lock(&song_mtx);
    pthread_mutex_lock(&scroll_mtx);
    if(title)
        title_len = strlen(title);
    if(title_len > 0) {
        if (do_scroll && title_len > titlebar_len) {
            int i, j = scroll_pos;
            s = malloc(sizeof(char) * (titlebar_len + 1));
            if(!s) {
                fprintf(stderr, "You are on VERY low memory!!!\n");
                goto end;
            }
            must_free = 1;
            for (i = 0; i < titlebar_len; i++, j++) {
                if(j >= title_len)
                    j -= title_len;
                s[i] = title[j];
            }
            s[i] = '\0';
            scroll_pos++;
            if (scroll_pos >= title_len) {
                scroll_pos = 0;
            }
        } else {
            s = title;
        }
    }
    pthread_mutex_unlock(&scroll_mtx);
    pthread_mutex_unlock(&song_mtx);
    
/*      DBGMSG("show_title: showing '%s'\n", s); */
    wgt_printf(B_TITLE, s);
end:
    if(must_free) free(s);
}

struct timeval* 
sum_timeval(struct timeval* res, struct timeval* a, struct timeval* b)
{
    res->tv_usec = a->tv_usec + b->tv_usec;
    res->tv_sec = a->tv_sec + b->tv_sec;
    if(res->tv_usec > 1000000) { /* handle overflow */
        res->tv_usec -= 1000000;
        res->tv_sec++;
    }
    return res;
}

static int is_scrolling = 0;

void scroll_handler(struct timeval* tv, void* param)
{ /* locks: display, song_mtx, scroll_mtx */
    if(show_msg_tit < 0) {
        show_title();
        XLockDisplay(display);
        XFlush(display);
        XUnlockDisplay(display);
    }

    if(do_scroll) {
        struct timeval now, next;
        /* calculate next scroll based on the *supposed* current time tv
           (if i used actual current time i'd be accumulating errors) */
        sum_timeval(&next, tv, &scroll_delay);
        /* ...unless it's VERY late (we've been stopped for a while) */
        gettimeofday(&now, NULL);
        if(now.tv_sec > next.tv_sec + scroll_delay.tv_sec + 1)
            sum_timeval(&next, &now, &scroll_delay);
        timer_schedule(&next, scroll_handler, param); /* param is unused */
        is_scrolling = 1;
        if (scroll_pos == 0 && !always_scroll)
            do_scroll = 0; /* this needs to be *after* timer_schedule */
    } else
        is_scrolling = 0;
}

void turn_on_scroll(void) /* locks: as scroll_handler (see above) */
{
    if (!do_scroll) {
        do_scroll = 1;
        scroll_pos = 0;
        if(!is_scrolling)
        {
            struct timeval tv;
            gettimeofday(&tv, NULL);
            scroll_handler(&tv, NULL); /* will call timer_schedule */
        }
    }
}


void show_list_number(unsigned long usec)
{
    int lstbar_len = btn_pos[B_LISTNAME].w / char_width;
    show_message(msg_listbar, usec,
         "%*d", lstbar_len, curr_song_num + 1);    
}


#if !HAVE_SNPRINTF
#error "need snprintf, get a C99 compiler (or write snprintf yourself)"
#endif

void draw_time(long min, long sec, int remain)
{
    struct coord* dpos = &btn_pos[B_DIR];
    int dirbar_len = dpos->w / char_width;
    char *buffer;
    int i /*, listnum*/;

    /* N.B. changing min to -min to show the - sign doesn't work if min == 0 */
/*  DBGMSG("show_time: showing time\n"); */
    if(remain)
        i = wgt_printf(B_LISTNAME, "-%ld:%02ld", min, sec);
    else
        i = wgt_printf(B_LISTNAME, "%02ld:%02ld", min, sec);
    /* hand draw the track number */
    if(dirbar_len > i + 1) {
        int len = dirbar_len - i;
        int x, w;
        buffer = malloc(len + 1);
        if(!buffer) {
            fprintf(stderr, "You are on VERY low memory!!!\n");
            return;
        }
        snprintf(buffer, len + 1, "%*d", len, curr_song_num + 1);
        x = dpos->x + char_width * i;
        w = dpos->w - char_width * i;
/*      DBGMSG("show_time: showing list number\n"); */
        draw_string(buffer, x, dpos->y, w);
/*      DBGMSG("show_time: redrawing...\n"); */
        RedrawRegion(x, dpos->y, w, dpos->h);
        free(buffer);
    }
}

void show_time(void) /* locks: display, song_mtx, player_mtx */
{
    float seconds = 0.0;
    long min, sec;
    song_pos pos;

    pthread_mutex_lock(&song_mtx);
    pthread_mutex_lock(&player_mtx);
    if(curr_player) {
        if(song_length.frames < 0 && show_time_remain)
            if(!curr_player->get_length(curr_player, &song_length))
                song_length.frames = -1;

        if(song_length.frames >= 0) {
            if(curr_player->get_pos(curr_player, &pos)) {
                if(show_time_remain)
                    seconds = song_length.seconds - pos.seconds;
                else
                    seconds = pos.seconds;
            }
        }
    }
    pthread_mutex_unlock(&player_mtx);
    pthread_mutex_unlock(&song_mtx);
/*  DBGMSG("show_time: pos = (%d,%f) len = (%d,%f)\n", pos.frames, pos.seconds, */
/*         song_length.frames, song_length.seconds); */

    min = (long)(seconds / 60);
    sec = (long)seconds - min * 60;
    draw_time(min, sec, show_time_remain);
}

/*
 * Progress bar functions
 */

void draw_progress_bar(int pos, int len)
{
    struct coord* pbpos = &btn_pos[B_PROGBAR];
    struct coord* pbzero = &btn_up[B_PROGBAR];
    struct coord* pbfull = &btn_down[B_PROGBAR];
    static long last_prog = -1;
    long progress = (pos * pbpos->w) / len;
/*  DBGMSG("draw_progress_bar: progress = %ld/%ld\n", progress, pbpos->w); */
    if(progress > pbpos->w) {
        DBGMSG("draw_progress_bar: position beyond the end!!!"
               " (resetting to %d)\n", pbpos->w);
        progress = pbpos->w;
    } else if(progress < 0) {
        DBGMSG("draw_progress_bar: position before beginning!!!"
               " (resetting to zero)\n");
        progress = 0;
    }

    if(progress != last_prog) {
        XLockDisplay(display);
        copyXPMArea(pbfull->x, pbfull->y, progress, pbfull->h,
                    pbpos->x, pbpos->y);
        copyXPMArea(pbzero->x + progress, pbzero->y, pbpos->w - progress,
                    pbzero->h, pbpos->x + progress, pbpos->y);
        RedrawRegion(pbpos->x, pbpos->y, pbpos->w, pbpos->h);
        XUnlockDisplay(display);
        last_prog = progress;
    }       
}

/* locks: display, song_mtx, player_mtx */
void show_progress_bar(void)
{
    int frames=0, total=1;
    pthread_mutex_lock(&song_mtx);
    pthread_mutex_lock(&player_mtx);
    if(curr_player) {
        if(song_length.frames < 0) /* invalid? */
            if(!curr_player->get_length(curr_player, &song_length))
                song_length.frames = -1; /* not needed really */
        total = song_length.frames;

        if(song_length.frames > 0) {
            song_pos pos;
            if(curr_player->get_pos(curr_player, &pos))
                frames = pos.frames;
            else
                frames = 0;
        }
    }
    pthread_mutex_unlock(&player_mtx);
    pthread_mutex_unlock(&song_mtx);
    draw_progress_bar(frames, total);
}

/*
 * toggles
 */

void toggle_scroll(void)
{
    if(!do_scroll)
        turn_on_scroll();
    else {
        scroll_pos = 0;
        do_scroll = 0;
        show_title(); /* show the title at pos. 0 */
    }      
}

enum { TOGGLE_MSG_TIME=1500000 }; /* show messages for 1.5sec  */

void random_on(void)
{
    show_message(msg_titlebar, TOGGLE_MSG_TIME, "random");
    random_flag = 1;
}

void random_off(void)
{
    show_message(msg_titlebar, TOGGLE_MSG_TIME, "no random");
    random_flag = 0;
}

void repeat_on(void)
{
    show_message(msg_titlebar, TOGGLE_MSG_TIME, "repeat");
    repeat_flag = 1;
}

void repeat_off(void)
{
    show_message(msg_titlebar, TOGGLE_MSG_TIME, "no repeat");
    repeat_flag = 0;
}

void time_remain_toggle(void)
{
    show_time_remain ^= 1;
    show_time();
}
