#ifndef thrcommH
#define thrcommH

/* Simple inline layer to enable threads (many sender and one receiver) to
 * communicate via a shared variable.
 */
#include <semaphore.h>

typedef struct thrchannel
{
    sem_t canput, canget;
    pthread_cond_t ack; /* signalled when the message has processed */
    pthread_mutex_t ack_mtx;
    void* volatile msg; /* shared variable */
    volatile int running; /* set to 1 if the thread is running, else 0 */
} thrchannel;

static inline void thrcomm_init(thrchannel* c)
{
    sem_init(&c->canput, 0, 1);
    sem_init(&c->canget, 0, 0);
    pthread_mutex_init(&c->ack_mtx, NULL);
    pthread_cond_init(&c->ack, NULL);
    c->msg = NULL;
    c->running = 0;
}

static inline void thrcomm_finish(thrchannel* c)
{
    sem_destroy(&c->canput);
    sem_destroy(&c->canget);
}

/* returns the message just read, and mark it as read.
   There MUST be a message, else the behaviour is undefined.
   (It will read the last one and discard the next at the moment).
   Use thrcomm_pending to test for this. This works based on the
   assumption that *only one* thread is listening. */
static inline void* thrcomm_recv(thrchannel* c)
{
    void* msg;
    sem_wait(&c->canget);
    msg = c->msg;
    sem_post(&c->canput);
    return msg;
}

/* wait for some messages */
static inline void thrcomm_wait(thrchannel* c)
{
    sem_wait(&c->canget);
    sem_post(&c->canget); /* reset the counter */
}

/* returns 1 (true) if the message was sent, 0 (false) otherwise.
   This, if called from the receiver thread, is likely to cause a deadlock.
   use thrcomm_trysend instead. */
static inline int thrcomm_send(thrchannel* c, void* msg)
{
    if(!c->running)
        return 0;
    sem_wait(&c->canput);
    if(!c->running) { /* receiver terminated while we were waiting */
        sem_post(&c->canput);
        return 0;
    }
    c->msg = msg;
    sem_post(&c->canget);
    return 1;
}

/* nonblocking send: returns true if the message was sent, 0 otherwise.
   if the message wasn't sent because the call would have blocked,
   errno is set to EAGAIN. */
static inline int thrcomm_trysend(thrchannel* c, void* msg)
{
    if(!c->running)
        return 0;
    if(sem_trywait(&c->canput) < 0)
        return 0; /* sem_trywait sets ERRNO to EAGAIN */
    if(!c->running) { /* receiver terminated while we were waiting */
        sem_post(&c->canput);
        return 0;
    }
    c->msg = msg;
    sem_post(&c->canget);
    return 1;
}

/* used by the receiver to say that it's able to receive messages */
static inline void thrcomm_listen(thrchannel* c)
{
    pthread_mutex_lock(&c->ack_mtx);
    c->running = 1;
    pthread_cond_broadcast(&c->ack);
    pthread_mutex_unlock(&c->ack_mtx);
}

/* used by the receiver to say that it won't be able to receive */
static inline void thrcomm_stop(thrchannel* c)
{
    pthread_mutex_lock(&c->ack_mtx);
    c->running = 0; /* important: must be BEFORE sem_post */
    if(sem_trywait(&c->canget) >= 0)
        sem_post(&c->canput);
    pthread_cond_broadcast(&c->ack);
    pthread_mutex_unlock(&c->ack_mtx);
}

/* waits until someone start listening */
static inline void thrcomm_waitlisten(thrchannel* c)
{
    pthread_mutex_lock(&c->ack_mtx);
    while(!c->running)
        pthread_cond_wait(&c->ack, &c->ack_mtx);
    pthread_mutex_unlock(&c->ack_mtx);
}

/* wait until the listener stops listening */
static inline void thrcomm_waitstop(thrchannel* c)
{
    pthread_mutex_lock(&c->ack_mtx);
    while(c->running)
        pthread_cond_wait(&c->ack, &c->ack_mtx);
    pthread_mutex_unlock(&c->ack_mtx);
}

/* these two 'ack' and 'waitack' requires that the message contained
   an address of a flag (tipically an automatic variable) which the
   receiver sets to something != 0 when it has processed the message. */
static inline void thrcomm_ack(thrchannel* c)
{
    pthread_mutex_lock(&c->ack_mtx);
    pthread_cond_broadcast(&c->ack);
    pthread_mutex_unlock(&c->ack_mtx);
}

static inline void thrcomm_waitack(thrchannel* c, volatile int* ackflag)
{
    pthread_mutex_lock(&c->ack_mtx);
    while(!*ackflag && c->running)
        pthread_cond_wait(&c->ack, &c->ack_mtx);
    pthread_mutex_unlock(&c->ack_mtx);
}

/* returns true if the receiver is running (=able to receive) */
static inline int thrcomm_running(thrchannel* c)
{
    return c->running;
}

/* returns true if there are some messages pending, false otherwise */
static inline int thrcomm_pending(thrchannel* c)
{
    int val;
    sem_getvalue(&c->canget, &val);
    return val; /* in this case only if val==0 there are messages */
}

#endif
