/*
 * wmdimp
 * Copyright (c) 2003 Daniele Tamino <aslo@wmdimp.cjb.net>
 *
 * Based on wmmp3: Copyright (c)1999 Patrick Crosby <xb@dotfiles.com>.
 *
 * This software covered by the GPL.  See COPYING file for details.
 *
 * utils.c
 *
 * various utility functions used by many functions
 *
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "utils.h"
#include "vect.h"
#include "parsecfg.h"
#include "debug.h"


/* get file extension */
char * getext(const char* fname)
{
    const char *p;
    const char *ext=0;
    for(p = fname; *p != '\0'; p++) {
        if(*p == '.')
            ext = p;
    }
    if(!ext)
        ext = p; /* empty string - p points to '\0' */
    return (char*)ext;
}

/* get filename (with ext) from complete path */
char * getname(const char* fname)
{
    const char *p;
    const char *name=0;
    for(p = fname; *p != '\0'; p++) {
        if(*p == '/')
            name = p + 1;
    }
    if(!name)
        name = fname;
    return (char*)name;
}

/* from Steven Jorgensen */
void stripspace(char *s) {
    char *t;
    t = s + strlen(s) - 1;
    while (t > s && isspace(*t)) {
        t--;
    }
    t++;
    *t = '\0';
}
 
static inline int push_char(struct vect* str, char c)
{
    if(!vect_incr(str, 1)) return 0;
    *_vect_get_last(str, char) = c;
    return 1;
}

/* simple string reader with ' and " style quoting and \ escaping,
 * term is a list of terminators, such as punctation charters, etc.
 * term may be NULL, so only spaces terminates the string.
 * term should not contain quoting (',") or escaping (\) characters.
 */
char* read_string(char* src, char** end, char* term)
{
    int quote = 0;
    int escape = 0;
    int done = 0;
    struct vect str;
    if(!vect_init(&str, sizeof(char))) return NULL;
    while(!done) {
        char c = *src++;
        if(escape) {
            if(!push_char(&str, c)) return NULL;
            escape = 0;
        } else if(quote && (c != '\\' || quote == '\'')) {
            if((int)c == quote)
                quote = 0;
            else
                if(!push_char(&str, c)) return NULL;
        } else {
            if(isspace(c) || c == '\0')
                done = 1;
            else {
                switch(c) {
                case '\'':
                case '"':   quote = (int)c; break;
                case '\\':  escape = 1; break;
                default:
                    if(term) {
                        char *t;
                        for(t = term; *t != '\0'; t++)
                            if(c == *t) {
                                done = 1;
                                break;
                            }
                    }
                    if(!done)
                        if(!push_char(&str, c)) return NULL;
                } /* end switch */
            } /* end else (c not a space)) */
        } /* end else (no quote or escaping) */
    } /* end main loop */
    if(!push_char(&str, '\0')) return NULL;
    if(end) *end = src;
    return (char*)str.v;
}

void free_str_list(str_list* l)
{
    str_list *n, *next;
    for(n = l; n != NULL; ) {
        next = n->next;
        if(n->str) free(n->str);
        free(n);
        n = next;
    }
}

void load_str_list(opt_tree_node* array, str_list** lst)
{
    opt_tree_node* n;
    str_list* l;
    if(*lst) {
        free_str_list(*lst);
        *lst = NULL;
    }
    if(!array) return;
    n = array->val.child;
    if(!n) return;
    l = NULL;
    assert(n->next); /* there MUST be at least one element */
    /* check if list is empty (only template), and use defaults in this case */
    if(n->next->next == NULL) {
        str_list* nw = (str_list*) malloc(sizeof(str_list));
        if(!nw) { fprintf(stderr, "VERY LOW MEM\n"); return; }
        nw->str = strdup(n->next->val.s);
        if(!nw->str) { fprintf(stderr, "VERY LOW MEM\n"); return; }
        nw->next = NULL;
        *lst = nw;
    } else {
        /* skip first (header) and last (template) elements */
        for(n = n->next; n->next; n = n->next) {
            str_list* nw = (str_list*) malloc(sizeof(str_list));
            if(!nw) { fprintf(stderr, "VERY LOW MEM\n"); return; }
            nw->str = strdup(n->val.s);
            if(!nw->str) { fprintf(stderr, "VERY LOW MEM\n"); return; }
            nw->next = NULL;
            if(l)
                l->next = nw;
            else
                *lst = nw;
            l = nw;
        }
    }
}

